#  """
#    Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import os
import time
from PIL import Image
import PySimpleGUI as sg
from wiliot_core import check_user_config_is_ok, WiliotDir, open_json
from wiliot_api import ManufacturingClient

class DataPullGUI:

    """
    To create .exe file for this script use the next line:
    pywiliot-testers> pyinstaller --onefile --windowed --add-data "./wiliot_testers/docs/wiliot_logo.png;./docs" ./wiliot_testers/utils/ppfp_tool.py
    """

    def __init__(self, owner_id='', single_crn='', output_dir=None):
        current_script = os.path.abspath(__file__)
        current_dir = os.path.dirname(current_script)
        os.chdir(current_dir)
        try:
            wiliot_logo = os.path.join('docs', 'wiliot_logo.png')
            if os.path.exists(wiliot_logo):
                pass
            else:
                raise Exception('Trying to locate .png')
        except Exception:
            wiliot_logo = os.path.join('..', 'docs', 'wiliot_logo.png')
        wiliot_logo_image = Image.open(wiliot_logo)
        wiliot_logo_image = wiliot_logo_image.resize((128, 50), Image.BICUBIC)
        wiliot_logo_image.save(wiliot_logo, format="png")
        sg.theme('GreenTan')
        self.layout = [
            [sg.Column([[sg.Image(wiliot_logo)]], justification='center')],
            [sg.Text('Owner ID', size=(15, 1)), sg.InputText(key='owner_id', default_text=owner_id)],
            [sg.Text('Environment'),
             sg.Combo(['Production', 'Test'], default_value='Production', key='env')],
            [sg.Text('Tester Type'),
             sg.Combo(['Offline', 'Sample'], default_value='Offline', key='tester_type')],
            [sg.Text('Select mode for Common Run Name Insert')],
            [sg.Radio('Single CRN', "RADIO1", default=True, key='single_crn', enable_events=True),
             sg.Radio('CRN List (CSV)', "RADIO1", key='csv', enable_events=True)],
            [sg.Column([[sg.Text('CRN', size=(15, 1)), sg.InputText(key='crn', default_text=single_crn)]], key='crn_col', visible=True),
             sg.Column([[sg.Text('CSV File', size=(15, 1)), sg.InputText(),
                         sg.FileBrowse(key='csv_file')]], key='csv_file_col', visible=False)],
            [sg.Text('Select Target Directory')],
            [sg.Text('Directory', size=(15, 1)),
             sg.InputText(output_dir if output_dir is not None else ''),
             sg.FolderBrowse(key='target_dir', initial_folder=output_dir)],
            [sg.Submit(), sg.Cancel()]
        ]

    def run(self):
        window = sg.Window('PPFP Tool', self.layout, finalize=True, size=(600, 350))

        while True:
            event, values = window.read()
            if event in (sg.WIN_CLOSED, 'Cancel'):
                break
            elif event == 'Submit':
                try:
                    env_dirs = WiliotDir()
                    config_dir_path = env_dirs.get_config_dir()
                    user_config_file_path = env_dirs.get_user_config_file()
                    loading_window = sg.Window('Loading', [[sg.Text('Loading...')]], finalize=True, no_titlebar=True,
                                               grab_anywhere=True)

                    if os.path.exists(user_config_file_path):
                        cfg_data = open_json(folder_path=config_dir_path, file_path=user_config_file_path)

                    window.minimize()
                    tester_type = 'offline-test' if values['tester_type'] == 'Offline' else 'sample-test'
                    file_path, api_key, is_successful = check_user_config_is_ok(env=values['env'][0:4].lower(),
                                                                                owner_id=values['owner_id'])
                    if not is_successful:
                        sg.popup_error('User configuration check failed')
                        continue
                    client = ManufacturingClient(api_key=api_key, env=values['env'][0:4].lower())
                    rsp = None

                    if values['single_crn']:
                        common_run_name_list = [values['crn'].strip()]
                    else:
                        common_run_name_list = []
                        with open(values['csv_file'], 'r') as f:
                            for line in f:
                                common_run_name_list.append(line.strip())
                    try:
                        for common_run_name in common_run_name_list:
                            out_file_path = os.path.join(values['target_dir'], f'{common_run_name}.zip')
                            with open(out_file_path, 'wb') as out_file:
                                rsp = client.get_file_for_ppfp(common_run_name, tester_type, out_file)
                    except Exception as e:
                        print(f'problem get file from cloud due to {e}')
                        loading_window.close()
                        sg.popup_error(f'An error occurred while loading CSV file - please check it', title='Error')
                        continue

                    loading_window.close()
                    if rsp:
                        print('Job Success')
                        done_window = sg.Window('Finish', [[sg.Text('Job Success')]], finalize=True, no_titlebar=True,
                                                grab_anywhere=True, auto_close=True, auto_close_duration=5)
                        time.sleep(5)
                        done_window.close()
                        window.close()
                    else:
                        sg.popup_error(f'An error occurred while getting data from cloud', title='Error')
                    break

                except Exception as e:
                    print(e)
                    sg.popup_error(f'An error occurred: {e}', title='Error')
                    break
            elif event == 'single_crn':
                window['csv_file_col'].update(visible=False)
                window['crn_col'].update(visible=True)
            elif event == 'csv':
                window['csv_file_col'].update(visible=True)
                window['crn_col'].update(visible=False)

        window.close()


if __name__ == '__main__':
    gui = DataPullGUI()
    gui.run()



import PySimpleGUI as sg
from wiliot_testers.test_equipment import *

try:
    main_sensor = YoctoSensor(None)
except Exception as ee:
    main_sensor = None
    raise Exception('No sensor is connected')

layout = [[sg.Text("Lightmeter Read")], [sg.Input(key='right_value')],
          [sg.Button('Calibrate'), sg.Button('Remove Calibrate'), sg.Button('Quit')]]

window = sg.Window('Lightmeter Read GUI', layout)
while True:
    event, values = window.read()
    if event == 'Calibrate':
        main_sensor.calibration(int(values['right_value']), True)
        break
    elif event == 'Remove Calibrate':
        main_sensor.calibration(0, False)
        break
    else:
        break
window.close()



"""
  Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

from wiliot_api import ManufacturingClient
from wiliot_core import check_user_config_is_ok, InlayTypes
from wiliot_testers.config.unusable_inlays import UnusableInlayTypes
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.calibration_test.calibration_test_by_tbp import *
from wiliot_testers.tester_utils import open_json, open_json_cache
from wiliot_testers.wiliot_tester_tag_result import *
from wiliot_testers.test_equipment import serial_ports
import serial.tools.list_ports
import PySimpleGUI as SimGUI
from PIL import Image
import re
import pathlib
import time

PRINT_FORMAT_TO_PASS_JOB_NAME = {'SGTIN': 'SGTIN_QR', 'Barcode': 'BARCODE_8', 'prePrint': 'BLANK'}
PASS_JOB_NUM = 2
TAG_COUNTER_DIGITS = 4


class ConfigDefaults(object):
    """
    contains the default values for the configuration json
    """

    def __init__(self):
        self.printer_defaults = {'TCP_IP': '192.168.6.61', 'TCP_PORT': '3003', 'TCP_BUFFER': '128',
                                 'enableLineSelection': 'Yes', 'enablePrinterAck': 'No',
                                 'printingDuringMovement': 'Yes'}
        self.single_band_gw_defaults = {'energizingPattern': '18', 'timeProfile': '3,7', 'txPower': '3',
                                        'rssiThreshold': '90', 'plDelay': '150'}
        self.dual_band_gw_defaults = {'energizingPattern': '18', 'secondEnergizingPattern': '52', 'timeProfile': '5,10',
                                      'txPower': '3', 'rssiThreshold': '90', 'plDelay': '100'}
        self.external_hw = {"temperatureSensorEnable": "No",
                            "AutoAttenuatorEnable": "No", "attnComport": "AUTO", "attnval": "0",
                            "scannerType": "rtscan"}

    def get_printer_defaults(self):
        return self.printer_defaults

    def get_single_band_gw_defaults(self):
        return self.single_band_gw_defaults

    def get_dual_band_gw_defaults(self):
        return self.dual_band_gw_defaults

    def get_external_hw_defaults(self):
        return self.external_hw


class DefaultGUIValues:
    def __init__(self, gui_type):
        if gui_type == 'Main':
            self.default_gui_values = {
                'toPrint': 'No', 'printOffset': '1', 'printingFormat': 'SGTIN',
                'batchName': 'test_reel',
                'tagGen': 'N/A',
                'testName': 'Ble_Test_Only', 'inlay': InlayTypes.TIKI_099.value,
                'desiredTags': '9999999', 'desiredPass': '9999999',
                'surface': SurfaceTypes.AIR.value,
                'conversion': ConversionTypes.STANDARD.value,
                'Environment': 'production', 'OwnerId': 'wiliot-ops', 'operator': '',
                'QRRead': 'No', 'QRcomport': 'COM3', 'QRoffset': '2',
                'QRtimeout': '200',
                'comments': '', 'maxFailStop': '50', 'maxYieldStop': '40',
                'pass_response_diff': '100', 'pass_response_diff_offset': '9999'}
        elif gui_type == 'Test':
            self.default_gui_values = {'passJobName': 'BARCODE_8', 'passJobNum': 2, 'sgtin': '',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': 'test', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        elif gui_type == 'SGTIN':
            self.default_gui_values = {'passJobName': 'SGTIN_QR', 'passJobNum': 2, 'sgtin': '(01)00850027865010(21)',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': '', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        elif gui_type == 'Barcode':
            self.default_gui_values = {'passJobName': 'BARCODE_8', 'passJobNum': 2, 'sgtin': '',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': '', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        elif gui_type == 'prePrint':
            self.default_gui_values = {'passJobName': 'BLANK', 'passJobNum': 2, 'sgtin': '',
                                       'stringBeforeCounter': '',
                                       'reelNumManually': '', 'firstPrintingValue': '0', 'tagLocation': '0',
                                       'tag_reel_location': '0'}
        else:
            self.default_gui_values = {}


def simple_calibration_gui():
    """
    open pop up window
    """

    dir_config = '../configs'
    tests_suites_configs_path = join(dir_config, 'tests_suites.json')
    with open(tests_suites_configs_path, 'r') as f:
        test_suite = json.load(f)

    inlay_type_list = [*test_suite]
    calibration_vals = []
    calibration_patterns = []
    calibration_powers = []
    calibration_timeprofiles = []
    layout = [[SimGUI.Text('Calibration Test Suite')],
              [SimGUI.Listbox(inlay_type_list, size=(35, len(inlay_type_list)), key='-CalibrationSuite-')],
              [SimGUI.Button('Ok')]]
    start_calib = True
    window = SimGUI.Window('Offline Simple Calibration', layout, default_element_size=(35, 1), auto_size_text=True,
                           auto_size_buttons=False,
                           default_button_element_size=(12, 1), element_justification='center')
    while True:  # the event loop
        event, values = window.read()
        if event == SimGUI.WIN_CLOSED:
            break
        if event == 'Ok':
            if values['-CalibrationSuite-']:  # if something is highlighted in the list
                for power_index in test_suite[values['-CalibrationSuite-'][0]]['tests']:
                    # if 'absGwTxPowerIndex' in power_index:
                    #     if int(power_index['absGwTxPowerIndex']) < 0 :
                    #         power_index['absGwTxPowerIndex'] = 19 + int(power_index['absGwTxPowerIndex'])
                    # else:
                    #     logging.warning('Problem with generating absGwTxPowerIndex value from\nTest suite, please check JSON file,\nWill calibrate with default power : 18')
                    #     power_index['absGwTxPowerIndex'] = 18
                    power_index['absGwTxPowerIndex'] = [14, 18]
                    if not 'timeProfile' in power_index:
                        power_index['timeProfile'] = [5, 15]
                    if not 'energizingPattern' in power_index:
                        logging.warning(
                            'Problem with read energizing power from test_suite \nPlease check test configuration file')
                        start_calib = False
                        break
                    calibration_vals.append(
                        {'pattern': power_index['energizingPattern'], 'power': power_index['absGwTxPowerIndex'],
                         'timeprofile': power_index['timeProfile']})
                    calibration_patterns.append(power_index['energizingPattern']) if power_index[
                                                                                         'energizingPattern'] not in calibration_patterns else calibration_patterns
                    # calibration_powers.append(power_index['absGwTxPowerIndex']) if power_index['absGwTxPowerIndex'] not in calibration_powers else calibration_powers
                    calibration_timeprofiles.append(power_index['timeProfile']) if power_index[
                                                                                       'timeProfile'] not in calibration_timeprofiles else calibration_timeprofiles
                calibration_powers.sort()
                calibration_power_range = range(power_index['absGwTxPowerIndex'][0],
                                                power_index['absGwTxPowerIndex'][-1] + 1)
                print('Starting calibration with values: powers: {}, time profile : {}, patterns : {}'.format(
                    str([calibration_power_range[0], calibration_power_range[-1]]),
                    str([calibration_timeprofiles[0][0], calibration_timeprofiles[0][-1]]), str(calibration_patterns)))
                SimGUI.popup(f"Calibration Started\n{values['-CalibrationSuite-'][0]}\nPlease wait",
                             button_type=SimGUI.POPUP_BUTTONS_NO_BUTTONS, non_blocking=True, no_titlebar=True,
                             auto_close=True, auto_close_duration=2)
                window.close()
                if start_calib:
                    top_score = start_calibration(
                        sweep_scan=[calibration_power_range[0], calibration_power_range[-1] + 1],
                        to_set=False, time_profiles_on=[calibration_timeprofiles[0][0]],
                        time_profiles_period=[calibration_timeprofiles[0][-1]],
                        energy_pattern_custom=calibration_patterns)

                    SimGUI.popup('Calibration Done\nWindow will be closed in 10 sec', keep_on_top=True,
                                 no_titlebar=True, auto_close=True, auto_close_duration=10)
                    print('Calibration Done')
                break
    window.close()
    return


def open_session(test_suite_list):
    """
    gets the user inputs from first GUI
    :return: dictionary of the values
    """

    def Collapsible(layout, key, title='', arrows=(SimGUI.SYMBOL_DOWN, SimGUI.SYMBOL_UP), collapsed=False):
        return SimGUI.Column(
            [[SimGUI.T((arrows[1] if collapsed else arrows[0]), enable_events=True, k=key + '-BUTTON-'),
              SimGUI.T(title, enable_events=True, key=key + '-TITLE-')],
             [SimGUI.pin(SimGUI.Column(layout, key=key, visible=not collapsed, metadata=arrows))]], pad=(0, 0))

    def check_structure(input_str):
        pattern = re.compile("^[A-Z0-9]{6}\.[0-9]{2}_[A-F]$")
        if not pattern.match(input_str):
            return False
        if not (0 < int(input_str[7:9]) <= 25):
            return False
        return True

    # move to r2r : return dict - r2r
    WILIOT_DIR = WiliotDir()
    dir_wiliot = WILIOT_DIR.get_wiliot_root_app_dir()
    tester_dir = join(dir_wiliot, 'offline')
    dir_config = join(tester_dir, '../configs')
    configuration_dir = os.path.join(dir_config, 'calibration_config.json')
    if os.path.exists(configuration_dir):
        with open(configuration_dir) as confile:
            configuration = json.load(confile)
            if configuration == '':
                confile.close()
                logging.warning('Setup configuration is not set, please execute calibration')
                with open(configuration_dir, 'w') as output:
                    configuration = default_calibration()
                    json.dump(configuration, output, indent=2, separators=(", ", ": "), sort_keys=False)

    else:
        logging.warning('Setup configuration is not set, please execute calibration')
        pathlib.Path(dir_config).mkdir(parents=True, exist_ok=True)
        with open(configuration_dir, 'w') as output:
            configuration = default_calibration()
            json.dump(configuration, output, indent=2, separators=(", ", ": "), sort_keys=False)

    current_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    parent_path = os.path.dirname(current_path)
    wiliot_logo = os.path.join(parent_path, "docs", "wiliot_logo.png")
    wiliot_logo_image = Image.open(wiliot_logo)
    wiliot_logo_image = wiliot_logo_image.resize((128, 50), Image.LANCZOS)
    wiliot_logo_image.save(wiliot_logo, format="png")

    folder_name = 'configs'
    file_name = 'gui_inputs_do_not_delete.json'
    gui_inputs_values = open_json(folder_path=folder_name, file_path=os.path.join(folder_name, file_name),
                                  default_values=DefaultGUIValues(gui_type='Main').default_gui_values)
    for key in DefaultGUIValues(gui_type='Main').default_gui_values.keys():
        if key not in gui_inputs_values.keys():
            gui_inputs_values[key] = DefaultGUIValues(gui_type='Main').default_gui_values[key]

    EXTEND_KEY = '-SECTION-'
    SimGUI.theme('GreenTan')
    inlay_group = tuple(x.value for x in InlayTypes if x.name not in UnusableInlayTypes.__members__)
    conversion_group = tuple([conv.value for conv in ConversionTypes])
    surface_group = tuple([surf.value for surf in SurfaceTypes])

    main_data_tab = [[SimGUI.Text('Reel_name:', size=(40, 1)), SimGUI.InputText(gui_inputs_values['batchName'],
                                                                                key='batchName')],
                     [SimGUI.Text('Operator:', size=(40, 1)), SimGUI.InputText(gui_inputs_values['operator'],
                                                                               key='operator')],
                     [SimGUI.InputCombo(('D'), visible=False, default_value='D3', key='tagGen')],
                     [SimGUI.Text('Test suite:', size=(40, 1)),
                      SimGUI.InputCombo(test_suite_list, default_value=gui_inputs_values['testName'],
                                        key='testName')],
                     [SimGUI.Text('Environment', size=(40, 1), visible=True),
                      SimGUI.InputCombo((tuple(['test', 'production'])), default_value=gui_inputs_values['Environment'],
                                        key='Environment', visible=True)],
                     [SimGUI.Text('Owner Id', size=(40, 1), visible=True),
                      SimGUI.InputCombo((tuple(['wiliot-ops', '852213717688'])),
                                        default_value=gui_inputs_values['OwnerId'], key='OwnerId', visible=True)],
                     [SimGUI.Text('', size=(40, 1))],
                     [SimGUI.Text('', size=(40, 1))],
                     [SimGUI.Text('Comments:', size=(40, 1)), SimGUI.InputText(gui_inputs_values['comments'],
                                                                               key='comments')]]

    reel_tab = [[SimGUI.Text('Inlay serial number (3 digits):', size=(40, 1), visible=True),
                 SimGUI.InputCombo(inlay_group, default_value=gui_inputs_values['inlay'], key='inlay', visible=True)],
                [SimGUI.Text('Surface:', size=(40, 1), visible=True),
                 SimGUI.InputCombo(surface_group,
                                   default_value=gui_inputs_values['surface'], key='surface',
                                   visible=True)],
                [SimGUI.Text('Conversion', size=(40, 1), visible=True),
                 SimGUI.InputCombo(conversion_group, default_value=gui_inputs_values['conversion'],
                                   key='conversion', visible=True)],
                [SimGUI.Text('External attenuator', size=(40, 1), visible=False),
                 SimGUI.Checkbox('', default=False, key='externalAttenuator',
                                 disabled=False, visible=False)],
                [SimGUI.Text('Attenuator ComPort: ', size=(40, 1), visible=False),
                 SimGUI.Spin(values=['Auto', 1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12, 13, 14, 15, 16, 17, 18],
                             initial_value=('Auto'),
                             size=(6, 1), key='attnComport', visible=False)],
                [SimGUI.Text('Attenuation value:', size=(40, 1), visible=False),
                 SimGUI.Spin(values=[i for i in range(5, 20)], key='attnval',
                             initial_value=5,
                             size=(6, 1), visible=False)]]

    print_tab = [[SimGUI.Text('To print?', size=(40, 1)),
                  SimGUI.InputCombo(('Yes', 'No'), default_value=gui_inputs_values['toPrint'], key='toPrint')],
                 [SimGUI.Text('What is the printing job format?', size=(40, 1)),
                  SimGUI.InputCombo(('Test', 'SGTIN', 'Barcode', 'prePrint'),
                                    default_value=gui_inputs_values['printingFormat']
                                    if gui_inputs_values['printingFormat'] != 'Test' else 'SGTIN',
                                    key='printingFormat')],
                 [SimGUI.Text('printing offset', size=(40, 1)),
                  SimGUI.Spin(values=[i for i in range(0, 5)], initial_value=int(gui_inputs_values['printOffset']),
                              size=(6, 1), key='printOffset')],
                 [SimGUI.Text('QR Validation:', size=(40, 1)),
                  SimGUI.InputCombo(('Yes', 'No'), visible=True, default_value=gui_inputs_values["QRRead"],
                                    key='QRRead')],
                 [SimGUI.Text('QR ComPort: ', size=(40, 1), visible=True),
                  SimGUI.Spin(values=[i for i in range(1, 15)], initial_value=(
                      int(gui_inputs_values['QRcomport'][-1]) if len(str(gui_inputs_values['QRcomport'])) > 1 else int(
                          gui_inputs_values['QRcomport'])),
                              size=(6, 1), key='QRcomport')],
                 [SimGUI.Text('QR offset (tags between GW and QR scanner):', size=(40, 1), visible=True),
                  SimGUI.Spin(values=[i for i in range(1, 10)], initial_value=int(gui_inputs_values['QRoffset']),
                              size=(6, 1), key='QRoffset')],
                 [SimGUI.Text('QR Timeout:', size=(40, 1), visible=True),
                  SimGUI.Spin(values=[i for i in range(100, 1000, 100)],
                              initial_value=int(gui_inputs_values['QRtimeout']),
                              size=(6, 1), key='QRtimeout')]
                 ]

    stop_conditiona_tab = [
        [SimGUI.Text('Ignore stop conditions', size=(40, 1)),
         SimGUI.Checkbox('', default=False, key='ignoreStop',
                         disabled=False, visible=True)],
        [SimGUI.Text('Max failed tags in a row:', size=(40, 1)),
         SimGUI.Spin(values=[i for i in range(1, 999999)], key='maxFailStop',
                     initial_value=int(
                         gui_inputs_values['maxFailStop']),
                     size=(6, 1))],
        [SimGUI.Text('Minimum yield [%]', size=(40, 1)),
         SimGUI.Spin(values=[i for i in range(1, 99)], key='maxYieldStop',
                     initial_value=int(
                         gui_inputs_values['maxYieldStop']),
                     size=(6, 1))],
        [SimGUI.Text('Desired amount of tags\n(will stop the run after this amount of tags):',
                     size=(40, 2)),
         SimGUI.InputText(gui_inputs_values['desiredTags'], key='desiredTags')],
        [SimGUI.Text('Desired amount of pass\n(will stop the run after this amount of passes):',
                     size=(40, 2)),
         SimGUI.InputText(gui_inputs_values['desiredPass'], key='desiredPass')]]

    layout = [
        [SimGUI.Image(filename=str(wiliot_logo))],
        [SimGUI.TabGroup(
            [[
                SimGUI.Tab('Main Run Data', main_data_tab),
                SimGUI.Tab('Reel Data', reel_tab),
                SimGUI.Tab('Stop Condition', stop_conditiona_tab),
                SimGUI.Tab('Printing', print_tab)]])],
        [SimGUI.Submit(button_color=('white', '#0e6251'), size=(10, 1)),
         SimGUI.Stretch(),
         SimGUI.Text('Version: {}'.format(get_version()), font='Helvetica 8 bold', justification='right')]]

    window = SimGUI.Window('Offline Tester', layout, font='Calibri')
    calib_val = False
    while True:
        event, values = window.read()

        if event is None:
            print('User exited the program')
            window.close()
            exit()

        elif event.startswith(EXTEND_KEY):
            window[EXTEND_KEY].update(visible=not window[EXTEND_KEY].visible)
            window[EXTEND_KEY + '-BUTTON-'].update(
                window[EXTEND_KEY].metadata[0] if window[EXTEND_KEY].visible else window[EXTEND_KEY].metadata[1])

        elif event == 'Submit':
            # make sure user input is legit
            reel_name_ver = False
            if int(values['printOffset']) != int(gui_inputs_values['printOffset']):
                # popup warning message:
                change_in_printing_offset = \
                    SimGUI.popup_ok_cancel('Are you sure you want to change the printing offset?\n'
                                           '(old value:{}, new value: {})?\n'.format(gui_inputs_values['printOffset'],
                                                                                     values['printOffset']),
                                           keep_on_top=True, font=('normal', 20))
                if change_in_printing_offset.lower() == 'cancel':
                    values['printOffset'] = gui_inputs_values['printOffset']
                    SimGUI.popup('printing offset was NOT change and equals to {}'.format(values['printOffset']),
                                 button_type='ok', font=14)

            if ' ' in values['batchName'] or '/' in values['batchName'] or '\\' in values['batchName']:
                print("Reel name could not contain spaces, '\\' or '/'\nplease fix it and press Submit")
                continue

            else:
                try:
                    int(values['desiredTags'])
                    int(values['desiredPass'])
                except Exception:
                    print("the following values should be numbers (integer): packetThreshold,"
                          "desiredTags, desiredPass (and should not be smaller than 2")
                    continue
                yes_no = ['Yes', 'No']
                if values['toPrint'] not in yes_no:
                    print("the following values should be 'Yes' or 'No': toPrint")
                    continue
                if values['Environment'] == 'production' and values['toPrint'].lower() == 'yes':
                    reel_name_ver = check_structure(values['batchName'])
                else:
                    reel_name_ver = True

                if reel_name_ver:
                    break
                else:
                    SimGUI.popup('Reel name error \n Reel name should be: <6 uppercase alphanumeric chars>.'
                                 '<2 integers minimum 01 maximum 25>_<upper case char minimum A maximum F>',
                                 button_type='ok', font=14)

        elif event == 'Advanced Calibration':
            # calib_val = True
            # break
            window.hide()
            calib_values = open_calibration_config()
            window.un_hide()

        elif event == 'Simple Calibration':
            window.hide()
            simple_calibration_gui()
            window.un_hide()

        elif event == SimGUI.WIN_CLOSED or event == 'Exit':
            print('User exited the program')
            break

    window.close()
    for key in gui_inputs_values.keys():
        if key not in values.keys():
            values[key] = gui_inputs_values[key]
    # to defend against user errors
    values['tagGen'] = values['tagGen'].upper()
    with open(os.path.join(folder_name, file_name), 'w') as f:
        json.dump(values, f, indent=2, separators=(", ", ": "), sort_keys=False)

    if calib_val:
        calib_values = open_calibration_config()

    return values


def get_printed_value(string_before_the_counter: str, first_counter: str, digits_in_counter=TAG_COUNTER_DIGITS):
    """
    builds the printed value
    :type string_before_the_counter: string
    :param string_before_the_counter: the sg1 Id of the tested reel
    :type digits_in_counter: int
    :param digits_in_counter: amount of digits in the tag counter field (usually 4)
    :type first_counter: string
    :param first_counter: counter of the run first tag
    :type printing_format: string
    :param printing_format: this run printing format (SGTIN, string)
    """
    first_print = str(string_before_the_counter)
    first_print += 'T'
    first_print += str(first_counter).zfill(digits_in_counter)
    is_ok = len(first_counter) <= digits_in_counter
    return first_print, is_ok


def update_tag_counter(current_tag_counter: str or int, value_to_add: int):
    return str(int(current_tag_counter) + int(value_to_add)).zfill(TAG_COUNTER_DIGITS)


def get_gui_inputs_values(printing_format):
    """
    A function that sends the GUI's input
    """
    folder_name = 'configs'
    file_name = get_print_user_config_file(printing_format)
    gui_inputs_values = open_json_cache(folder_path=folder_name, file_path=os.path.join(folder_name, file_name),
                                        default_values=DefaultGUIValues(printing_format).default_gui_values)
    return gui_inputs_values


def printing_process_of_test_and_sgtin(window, printing_format, gui_inputs_values, is_new_reel=True):
    """
    A function that does all the work of the printing process
    """
    reel_number = ''
    pass_job_name = None
    is_ok = True
    while True:
        event, values = window.read()
        if printing_format != 'Test':
            pass_job_name = values['passJobName']
        if event == SimGUI.WINDOW_CLOSED or event is None:
            is_ok = False
            break

        if event == 'Submit':
            try:
                if printing_format == 'prePrint':
                    if 'T' not in values['firstFullExternalId']:
                        window['-OUTPUT-'].update(
                            'scanned external id must contains the char T')
                        continue
                    external_id_list = values['firstFullExternalId'].split('T')
                    if len(external_id_list) != 2:
                        window['-OUTPUT-'].update(
                            'scanned external id must be: <REEL ID>T<TAG COUNT>')
                        continue
                    if len(external_id_list[1]) != 4:
                        window['-OUTPUT-'].update(
                            'Counter should be equal to 4 digits')
                        continue
                    values['firstPrintingValue'] = external_id_list[1]
                    values['reelNumManually'] = external_id_list[0]

                if len(str(values['firstPrintingValue'])) > TAG_COUNTER_DIGITS:
                    window['-OUTPUT-'].update(
                        f"you entered: {values['firstPrintingValue']} but length must be {TAG_COUNTER_DIGITS}"
                    )
                    continue
                if printing_format == 'SGTIN' or 'SGTIN' in values['passJobName']:
                    reel_number = str(values['sgtinNumManually'])
                    if not len(str(values['sgtinNumManually'])) == 22:
                        window['-OUTPUT-'].update(
                            f'SGTIN number is not equal to 22 chars!!\n'
                            f'The current SGTIN length is: {len(values["sgtinNumManually"])}')
                        continue
                    if not len(str(values['reelNumManually'])) == 4:
                        window['-OUTPUT-'].update(
                            'Reel number is not equal to 4 chars!!\n'
                            'Please enter correct Reel number')
                        continue
                if (printing_format == 'Barcode' or 'Barcode' in values['passJobName']) and \
                        len(values['reelNumManually']) != 3:
                    window['-OUTPUT-'].update(
                        'Reel number is not equal to 3 chars!!\n'
                        'Please enter correct Reel number')
                    continue

                pass_job_name = values['passJobName']

                reel_number += str(values['reelNumManually'])
                first_counter = values['firstPrintingValue']
                first_print, is_ok = get_printed_value(reel_number, first_counter)
                window['-OUTPUT-'].update('The first tag printing value will be:\n' + first_print)
            except Exception as e:

                window['-OUTPUT-'].update(f'got exception during parsing, please try again: {e}')
                continue

            if printing_format == 'prePrint':
                reel_number = str(values['reelNumManually'])
                first_print, is_ok = get_printed_value(reel_number, values['firstPrintingValue'])
                if is_new_reel:
                    msg = f'Are you sure the first tag for testing is:\n\n' \
                          f'      {first_print}\n'
                else:
                    msg = f'Are you sure the tag above the coupler is:\n\n' \
                          f'      {first_print}\n\n'
                    values['firstPrintingValue'] = update_tag_counter(values['firstPrintingValue'], 1)
                    first_print, _ = get_printed_value(reel_number, values['firstPrintingValue'])
                    msg += f'it means that the first tag for testing is:\n\n' \
                           f'      {first_print}\n'
                sure_to_submit = SimGUI.popup_yes_no(msg,
                                                     title='first tag for testing', font=('normal', 18),
                                                     keep_on_top=True)
                if sure_to_submit.lower() == 'no':
                    window['-OUTPUT-'].update('')
                    continue
                else:
                    window['-OUTPUT-'].update('The first tag printing value will be:\n' + first_print)

        break

    v = {'passJobName': pass_job_name, 'digitsInCounter': TAG_COUNTER_DIGITS, 'passJobNum': PASS_JOB_NUM,
         'firstPrintingValue': values['firstPrintingValue'], 'tagLocation': values['firstPrintingValue'],
         'tag_reel_location': gui_inputs_values['tag_reel_location']}

    data_to_save = {'passJobName': pass_job_name,
                    'passJobNum': PASS_JOB_NUM,
                    'sgtin': values['sgtinNumManually'] if 'sgtinNumManually' in values else '',
                    'reelNumManually': values['reelNumManually'],
                    'firstPrintingValue': values['firstPrintingValue'],
                    'tagLocation': values['firstPrintingValue'],
                    'tag_reel_location': gui_inputs_values['tag_reel_location']}

    if printing_format == 'Test':
        v['stringBeforeCounter'] = \
            values['sgtinNumManually'] if 'sgtinNumManually' in values else '' + values['reelNumManually']
        v['failJobName'] = pass_job_name
    else:
        v['stringBeforeCounter'] = reel_number
        v['failJobName'] = 'line_'
        v['printingFormat'] = printing_format
        data_to_save['stringBeforeCounter'] = reel_number
        data_to_save['printingFormat'] = printing_format

    folder_name = 'configs'
    file_name = get_print_user_config_file(printing_format)
    f = open(os.path.join(folder_name, file_name), "w")
    json.dump(data_to_save, f)
    f.close()
    return v, is_ok


def get_print_user_config_file(printing_format):
    if printing_format.lower() == 'test':
        filename = 'gui_printer_inputs_4_Test_do_not_delete.json'
    elif printing_format in PRINT_FORMAT_TO_PASS_JOB_NAME.keys():
        filename = 'gui_printer_inputs_4_SGTIN_do_not_delete.json'
    else:
        raise Exception(f'unsupported printing format: {printing_format}, valid formats are: sgtin, barcode, test')
    return filename


def printing_test_window():
    """
    opens the GUI for user input for test print
    :return: dictionary of user inputs
    """
    printing_format = 'Test'
    gui_inputs_values = get_gui_inputs_values(printing_format)

    if gui_inputs_values['reelNumManually'] == "":
        reel_num = 'test_test_test_test_X_test'
        gui_inputs_values['sgtin'] = reel_num[:22]
        gui_inputs_values['reelNumManually'] = reel_num[22:]
        gui_inputs_values['firstPrintingValue'] = '0'
        gui_inputs_values['tagLocation'] = '0'
        gui_inputs_values['tag_reel_location'] = '0'

    layout = [[SimGUI.Text('Job to print for pass and fail:'),
               SimGUI.InputCombo(('SGTIN_only', 'SGTIN_QR', 'BARCODE_8', 'devkit_TEO', 'devkit_TIKI', 'empty'),
                                 default_value=gui_inputs_values['passJobName'], key='passJobName')],
              [SimGUI.Text('pass line number:'),
               SimGUI.Input(gui_inputs_values['passJobNum'], key='passJobNum')],
              [SimGUI.Text("What is the first counter number?")],
              [SimGUI.Input(gui_inputs_values['firstPrintingValue'], key='firstPrintingValue')],
              [SimGUI.Checkbox('Insert reel number manually?', default=False, key='isManually')],
              [SimGUI.Text("For manual mode only - what is the sgtin number?", key='sgtinNumManuallyText'),
               SimGUI.Input(gui_inputs_values['sgtin'], key='sgtinNumManually')],
              [SimGUI.Text("For manual mode only - what is the reel number?", key='reelNumManuallyText'),
               SimGUI.Input(gui_inputs_values['reelNumManually'], key='reelNumManually')],
              [SimGUI.Text(size=(60, 3), key='-OUTPUT-')],
              [SimGUI.Text("*Tags in this run will not be serialized")],
              [SimGUI.Button('Check first print'), SimGUI.Button('Submit', button_color=('white', '#0e6251'))]]
    window = SimGUI.Window('Printing Test', layout, keep_on_top=True)
    wanted_v, wanted_is_ok = printing_process_of_test_and_sgtin(window, printing_format, gui_inputs_values)
    window.close()
    return wanted_v, wanted_is_ok


def printing_sgtin_window(env='', owner_id='wiliot-ops', printing_format='SGTIN'):
    """
    opens the GUI for user input for SGTIN print
    :return: dictionary of user inputs
    """
    read_only = False
    gui_inputs_values = get_gui_inputs_values(printing_format)

    # Checking if it's a new reel run and updating the GUI inputs according to this
    new_run = SimGUI.popup_yes_no('    New Reel?    \n', keep_on_top=True, font=('normal', 20))
    if env == 'production' or env == '':
        env = 'prod'
    if new_run == 'Yes':
        gui_inputs_values['tagLocation'] = '0'
        gui_inputs_values['tag_reel_location'] = '0'
        if printing_format != 'prePrint':
            try:
                logging.info('Receiving data from the cloud, please wait')
                reel_num = get_reel_name_from_cloud_api(env, owner_id, printing_format)
                print(reel_num)
            except Exception:
                logging.warning('Problem with receiving data from cloud')
                raise Exception
        else:
            reel_num = ''
        gui_inputs_values['firstPrintingValue'] = '0000'
        if 'data' in reel_num:
            reel_number = reel_num['data']
            if printing_format == 'SGTIN':
                gui_inputs_values['sgtin'] = reel_number[:22]
                gui_inputs_values['reelNumManually'] = reel_number[22:26]
            elif printing_format == 'Barcode':
                gui_inputs_values['sgtin'] = ''
                gui_inputs_values['reelNumManually'] = reel_number
            else:
                raise Exception(f'printing_format is not supported: {printing_format}')
            read_only = True
        else:
            gui_inputs_values['sgtin'] = ''
            gui_inputs_values['reelNumManually'] = ''
            read_only = False

    layout = [
        [SimGUI.Text(printing_format, font=('normal', 16))],
        [SimGUI.Text('Job to print for pass:'),
         SimGUI.InputCombo(('SGTIN_only', 'SGTIN_QR', 'BARCODE_8', 'devkit_TEO', 'devkit_TIKI', 'empty'),
                           default_value=PRINT_FORMAT_TO_PASS_JOB_NAME[printing_format],
                           key='passJobName')]]
    if printing_format == 'prePrint':
        scan_str = 'place the first tag before the coupler and scan it' if new_run == 'Yes' else \
            'scan the tag above the coupler'
        layout += [
            [SimGUI.Text(scan_str, font=('normal', 14), text_color='red')],
            [SimGUI.Input("", key='firstFullExternalId')]]
    else:
        layout += [
            [SimGUI.Text("First counter number")],
            [SimGUI.Input(0000 if read_only else gui_inputs_values['firstPrintingValue'], key='firstPrintingValue',
                          readonly=read_only)],
            [SimGUI.Text("Reel number", key='reelNumManuallyText'),
             SimGUI.Input(gui_inputs_values['reelNumManually'], key='reelNumManually', readonly=read_only)]]
    if printing_format == 'SGTIN':
        layout += [
            [SimGUI.Text("SGTIN number", key='sgtinNumManuallyText'),
             SimGUI.Input(gui_inputs_values['sgtin'], key='sgtinNumManually', readonly=read_only)]]
    layout += [
        [SimGUI.Text(size=(60, 3), key='-OUTPUT-', text_color='red')],
        [SimGUI.Text('Tag Reel Location: {}'.format(gui_inputs_values['tag_reel_location']))],
        [SimGUI.Button('Submit', button_color=('white', '#0e6251'))]]
    window = SimGUI.Window('Printing Setting', layout, keep_on_top=True)
    wanted_v, wanted_is_ok = printing_process_of_test_and_sgtin(window, printing_format, gui_inputs_values,
                                                                new_run == 'Yes')
    window.close()
    return wanted_v, wanted_is_ok


def save_screen(tested, passed, yield_, missing_labels, problem_in_locations_hist_val, ttfgp_avg, ttfp_max_error=1,
                default_upload_value=None, responded=None):
    """
    open last GUI
    :type tested: int
    :param tested: amount of tested tags
    :type passed: int
    :param passed: amount of passed tags
    :type yield_: float
    :param yield_: yield in the run
    :type missing_labels: int
    :param missing_labels: amount of missing_labels tags
    :type problem_in_locations_hist_val: dictionary or None
    :param problem_in_locations_hist_val: histogram of problem in the run (amount of locations)
    :type ttfgp_avg: float or None
    :param ttfgp_avg: average of ttfgp (time to first good packet) in this run
    :type default_upload_value: string ('Yes' or 'No')
    :param default_upload_value: default value to use in upload to cloud field
    :return dictionary with the user inputs (should upload, last comments)
    """
    SimGUI.theme('GreenTan')
    if ttfgp_avg is None or str(ttfgp_avg) == 'nan':
        ttfgp_avg_line = []
    elif ttfgp_avg < ttfp_max_error:
        ttfgp_avg_line = [
            SimGUI.Text("Average time to first good packet is OK (" + '{0:.4g}'.format(ttfgp_avg) + " secs)",
                        border_width=10, background_color='green')]
    else:
        ttfgp_avg_line = [SimGUI.Text("Average time to first good packet is too high (" + '{0:.4g}'.format(ttfgp_avg) +
                                      " secs)", border_width=10, background_color='red')]
    if default_upload_value is not None:
        def_upload = default_upload_value
    else:
        def_upload = "Yes"

    layout = [
        [SimGUI.Text('Tags tested = ' + str(tested), size=(21, 1)),
         SimGUI.Text('Tags responded = ' + str(responded if responded is not None else 0), size=(21, 1)),
         SimGUI.Text('Tags passed = ' + str(passed), size=(21, 1))],
        [SimGUI.Text('Yield = ' + '{0:.4g}'.format(yield_) + '%', size=(21, 2)),
         SimGUI.Text('Missing labels = ' + str(missing_labels), size=(21, 2))],
        ttfgp_avg_line,
        [SimGUI.Text('Would you like to upload this log to the cloud?'),
         SimGUI.InputCombo(('Yes', 'No'), default_value=def_upload, key='upload')],
        [SimGUI.Text('Post run comments:')],
        [SimGUI.InputText('', key='comments')],
        [SimGUI.Submit(button_color=('white', '#0e6251'))]]

    window = SimGUI.Window('Offline Tester - Summary ', layout, force_toplevel=True, keep_on_top=True)
    while True:
        event, values = window.read()
        # See if user wants to quit or window was closed
        if event == SimGUI.WINDOW_CLOSED or event is None:
            print("user exited the program, upload did not happen")
            break
        if event == 'Submit':
            break
    window.close()
    # sys.exit(0)
    return values


def get_reel_name_from_cloud_api(env, owner_id, printing_format='SGTIN'):
    """
    API to receive reel number from cloud (should use it to avoid duplications).
    :return: the reel number (in 0x)
    """
    assert ('R2R_station_name' in os.environ), 'R2R_station_name is missing from PC environment variables'
    tester_station_name = os.environ['R2R_station_name']

    try:
        file_path, api_key, is_successful = check_user_config_is_ok(env=env, owner_id=owner_id)
        client = ManufacturingClient(api_key=api_key, logger_=logging.getLogger().name, env=env)
        payload = {"printerId": tester_station_name}
        reel_id_3_char = printing_format == "Barcode"
        return client.get_reel_id(owner_id, payload, reel_id_3_char)

    except Exception as e:
        raise Exception(f"An exception occurred at get_reel_name_from_cloud_API: {e}")


"""
    R2R Arduino
"""

ARDUINO_BAUD_RATE = 1000000


class R2rGpio(object):
    """
    class to open and use communication to Arduino on R2R machine
    """

    def __init__(self, logger_name=None):
        """
        initialize params and port
        """
        self.baud_rate = ARDUINO_BAUD_RATE
        self.comport = ''
        self.s = None
        self.connected = False
        self.logger = logging.getLogger(logger_name) if logger_name is not None else None
        ports_list = serial_ports()
        if len(ports_list) == 0:
            raise Exception("no serial ports were found. please check your connections")
        self.connect(ports_list)

    def is_connected(self):
        if self.s is None:
            return False

        if not self.s.isOpen():
            return False

        response = self.query("*IDN?")

        if "Williot R2R GPIO" in response:
            self.connected = True
            if self.logger is None:
                print('R2R: Found ' + response + " Serial Number " + self.query("SER?"))
            else:
                self.logger.info('R2R: Found ' + response + " Serial Number " + self.query("SER?"))
            self.s.flushInput()
            return True
        else:
            self.s.close()
            return False

    def connect(self, ports_list):
        for port in ports_list:
            try:
                self.comport = port
                self.s = serial.Serial(self.comport, self.baud_rate, timeout=0, write_timeout=0)

                if self.is_connected():
                    break

            except (OSError, serial.SerialException):
                pass
            except Exception as e:
                if self.logger is None:
                    print(e)
                else:
                    self.logger.warning(f'R2R: connect: {e}')
        if not self.connected:
            raise Exception('Could NOT connect to the Arduino, please check connections')

    def __del__(self):
        if self.s is not None:
            self.s.close()

    def write(self, cmd):
        """
        Send the input cmd string via COM Socket
        """
        if self.s.isOpen():
            pass
        else:
            self.s.open()

        try:
            self.s.flushInput()
            self.s.write(str.encode(cmd))
        except Exception:
            pass

    def query(self, cmd):
        """
        Send the input cmd string via COM Socket and return the reply string
        :return: massage from arduino (w/o the '\t\n')
        """
        if self.s.isOpen():
            pass
        else:
            self.s.open()
            sleep(1)
        self.s.flushInput()
        sleep(1)
        try:
            self.s.write(str.encode(cmd))
            sleep(2)
            data = self.s.readlines()
            value = data[0].decode("utf-8")
            # Cut the last character as the device returns a null terminated string
            value = value[:-2]
        except Exception:
            value = ''
        return value

    def read(self):
        """
        Send the input cmd string via COM Socket and return the reply string
        :return: massage from arduino (w/o the '\t\n')
        """
        if self.s.isOpen():
            pass
        else:
            self.s.open()
        try:
            timeout = time.time() + 5  # 5 second to receive arudio pulse
            while self.s.in_waiting == 0 and time.time() < timeout:
                pass
            if time.time() > timeout:
                logging.warning('No data received from Arduino')
                self.reconnect()

            data = self.s.readlines()
            self.s.flushInput()
            value = data[0].decode("utf-8")
            # Cut the last character as the device returns a null terminated string
            value = value[:-2]
        except Exception:
            value = ''
        return value

    def gpio_state(self, gpio, state):
        """
        gets the gpio state:
            my_gpio.gpio_state(3, "ON")
               start"on"/stop"off"
            my_gpio.gpio_state(4, "ON")
               enable missing label
        :param gpio: what gpio to write to
        :param state: to what state to transfer (ON / OFF)
        :return: reply from Arduino
        """
        cmd = 'GPIO' + str(gpio) + "_" + state
        replay = self.query(cmd)
        return replay

    def pulse(self, gpio, time):
        """
        send a pulse to the r2r machine:
            my_gpio.pulse(1, 1000)
               Pass
            my_gpio.pulse(2, 1000)
               fail
        :param gpio: what gpio to write to
        :param time: how long is the pulse
        :return: True if succeeded, False otherwise
        """
        cmd = 'GPIO' + str(gpio) + '_PULSE ' + str(time)
        self.write(cmd)
        sleep(time * 2 / 1000)
        replay = self.read()
        if replay == "Completed Successfully":
            return True
        else:
            return False

    def reconnect(self):
        if self.is_connected():
            logging.info(f'Already connected to Arduino on {self.comport}')
            return True
        try:
            # Attempt to open a connection to the specified comport
            self.s = serial.Serial(self.comport, self.baud_rate, timeout=0, write_timeout=0)
            logging.info(f'Connection to Arduino on {self.comport} reestablished.')
            reconnect_success = True

        except serial.serialutil.SerialException:
            logging.warning(
                f'Connection to {self.comport} failed. Attempting to close the connection and detect Arduino...')
            try:
                self.s.close()
                time.sleep(1)
                self.s = serial.Serial(self.comport, self.baud_rate, timeout=0, write_timeout=0)
                logging.info(f'Connection to Arduino on {self.comport} reestablished.')
                reconnect_success = True

            except Exception as e:
                logging.warning(f"Failed to close connection to {self.comport} {e} - Please restart connection")
                return False

        if reconnect_success:

            if self.is_connected():
                logging.info(
                    'Connection was restablished on COM{} with version {}'.format(self.comport, self.query("SER?")))
                self.s.flushInput()
            else:
                self.s.close()
                logging.warning(f"Failed to close connection to {self.comport} - Please restart connection")
                reconnect_success = False
        return reconnect_success


if __name__ == '__main__':
    a = printing_sgtin_window(env='prod', owner_id='wiliot-ops', printing_format='prePrint')
    rsp = get_reel_name_from_cloud_api(env='test', owner_id='wiliot-ops', printing_format='Barcode')

    open_session(['Single Band', 'Dual Band'])
    printing_test_window()
    printing_sgtin_window(env='test', printing_format='Barcode')
    printing_sgtin_window(env='test', printing_format='SGTIN')

    r = R2rGpio()
    open_session(['Single Band', 'Dual Band'])
    pass



import serial
import time
import threading
from time import sleep
import serial.tools.list_ports
import logging
import os
from datetime import datetime


class CountThread(object):
    """
    Counting the number of tags
    """

    def __init__(self, stop_event, rows=1, max_movement_time=18, ther_cols=1):
        self.max_movement_time = max_movement_time
        self.available_ports = [s.device for s in serial.tools.list_ports.comports()]
        self.get_ports_of_arduino()
        self.baud = 1000000
        self.ports = self.get_ports_of_arduino()
        self.logger_file = self.setup_logger()
        try:
            self.comPortObj = serial.Serial(self.ports[0], self.baud, timeout=0.1)
            sleep(1)
            self.config()
        except serial.SerialException:
            self.logger_file.error("NO ARDUINO")
            raise Exception('could not connect to the Arduino')
        self.rows = rows
        self.ther_cols = ther_cols
        self.stop = stop_event
        self.tested = 0

    def setup_logger(self):
        user_home = os.path.expanduser("~")
        now = datetime.now().strftime('%m/%d/%Y %H:%M:%S')
        log_path = os.path.join(user_home, 'Downloads', f'{now}.log')
        logging.basicConfig(filename=log_path, level=logging.INFO,
                            format='%(asctime)s - %(levelname)s - %(message)s')
        return logging.getLogger()

    def config(self):
        """
        @return:
        @rtype:
        """
        self.comPortObj.write(f'*time_btwn_triggers {int(self.max_movement_time)}'.encode())
        rsp = self.comPortObj.readline()
        if rsp == b'' or rsp != b'No pulses yet...\r\n':
            raise Exception('A problem in the first message of Arduino')
        for i in range(5):
            sleep(0.500)
            rsp = self.comPortObj.readline()
            if rsp.decode() == f'Max movement time was set to {int(self.max_movement_time)}[sec]\r\n':
                self.logger_file.info(f'config Arduino and got the following msg: {rsp.decode()}')
                return
        raise Exception('Arduino Configuration was failed')

    def get_ports_of_arduino(self):
        """
        Gets all the ports we have, then chooses Arduino's ports
        """
        arduino_ports = []
        for p in serial.tools.list_ports.comports():
            if 'Arduino' in p.description:
                arduino_ports.append(p.device)
        if not arduino_ports:
            self.logger_file.info('NO ARDUINO')
        return arduino_ports

    def run(self):
        """
        Tries to read data and then counts the number of tags
        """
        while not self.stop.is_set():
            time.sleep(0.100)
            data = ''
            try:
                data = self.comPortObj.readline()
            except Exception as ee:
                self.logger_file.error(f"NO READLINE: {ee}")
            if data.__len__() > 0:
                try:
                    tmp = data.decode().strip(' \t\n\r')
                    if "pulses detected" in tmp:
                        self.tested += (self.rows * int(self.ther_cols))
                    self.logger_file.info(f'Arduino msg: {tmp}')

                except Exception as ee:
                    self.logger_file.error(f'Warning: Could not decode counter data or Warning: {ee}')
                    continue
        self.comPortObj.close()

    def get_tested(self):
        """
        returns the number of tags
        """
        return self.tested


def run_count_thread():
    stop_event = threading.Event()

    class Logger:
        def error(self, msg):
            self.logger_file.info(f'error: {msg}')

        def info(self, msg):
            self.logger_file.info(f"msg: {msg}")

    logger = Logger()

    rows_number = 1
    max_movement_time = 18
    thermodes_col = 1

    count_process = CountThread(stop_event, rows_number, max_movement_time, thermodes_col)
    count_process_thread = threading.Thread(target=count_process.run)

    count_process_thread.start()

    prev_tested_count = 0

    try:
        while True:
            current_tested_count = count_process.get_tested()

            if current_tested_count != prev_tested_count:
                print(f"Number of tags tested: {current_tested_count}")
                prev_tested_count = current_tested_count

            time.sleep(1)

    except KeyboardInterrupt:
        stop_event.set()
        count_process_thread.join()
        print("\nStopped counting and exited gracefully.")


if __name__ == "__main__":
    run_count_thread()



csv_dictionary = {
    'TEO_086': {'inlay': 'TEO_086', 'number': '086', 'number_of_rows': 1, 'received_channel': '37',
                'energy_pattern_val': 18, 'time_profile_val': [5, 40]},
    'TIKI_096': {'inlay': 'TIKI_096', 'number': '096', 'number_of_rows': 2, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40]},
    'TIKI_099': {'inlay': 'TIKI_099', 'number': '099', 'number_of_rows': 3, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40]},
    'BATTERY_107': {'inlay': 'BATTERY_107', 'number': '107', 'number_of_rows': 4, 'received_channel': '37',
                    'energy_pattern_val': 17, 'time_profile_val': [5, 40]},
    'TIKI_117': {'inlay': 'TIKI_117', 'number': '117', 'number_of_rows': 5, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40]},
    'TIKI_121': {'inlay': 'TIKI_121', 'number': '121', 'number_of_rows': 6, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40]},
    'TIKI_122': {'inlay': 'TIKI_122', 'number': '122', 'number_of_rows': 7, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40]},
    '': {'inlay': '', 'number': '', 'number_of_rows': 0, 'received_channel': '', 'energy_pattern_val': 0,
         'time_profile_val': [0, 0]}}



"""
  Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
import copy
from yoctopuce.yocto_temperature import *
import http.client
import mimetypes
from codecs import encode
import threading
import logging
import traceback
import json
import serial.tools.list_ports
import time
import pandas as pd
import os
from csv import writer, DictWriter
import ast
from threading import Event
import numpy as np
from queue import Empty
import sys
from os.path import dirname
import PySimpleGUI as SimGUI
import pathlib
from wiliot_api import ManufacturingClient, WiliotCloudError
from wiliot_testers.wiliot_tester_tag_result import *
from wiliot_core import check_user_config_is_ok
# TODO: consider to use for cross platform directories


###########################################
#            Tester Management            #
###########################################

class TesterName(Enum):
    """
    determines tester type (affects output file format set in class CsvLog)
    """
    OFFLINE = 'offline'
    TAL15K = 'tal15k'
    CONVERSION = 'conversion'
    SAMPLE = 'sample'
    NONE = ''


class HeaderType(Enum):
    """
    determines which output file is generated in class CsvLog (run data or tags data)
    """
    TAG = 'tag'
    RUN = 'run'
    PACKETS = 'packets'
    NONE = ''


def get_ports(device_name):
    """
    search the com ports for the desired device (arduino/ gw)
    :type device_name: string
    :param device_name: arduino or gw (gateway)
    :return: ports list of the desired devices
    """
    if device_name == 'gw' or device_name == 'gateway':
        str_device = 'Wiliot GW'
        baud = 921600
    elif device_name == 'arduino':
        str_device = 'Wiliot Tester GPIO unit'
        baud = 1000000
    else:
        print("please selecet a valid device names: {'gw','arduino'}")
        return []
    ports = list(serial.tools.list_ports.comports())
    device_ports = []
    for p in ports:
        try:
            s = serial.Serial(port=p.device, baudrate=baud, timeout=1)
            if s.isOpen():
                s.write(b'*IDN?')
                current_device = s.readline()
                if str_device in current_device.decode():
                    device_ports.append(p.device)
                    print(current_device.decode())
                s.close()
        except (OSError, serial.SerialException):
            pass
        except Exception as e:
            raise (e)
    return device_ports

def dict_to_csv(dict_in, path, append=False, only_titles=False):
    if append:
        method = 'a'
    else:
        method = 'w'
    with open(path, method, newline='') as f:
        dict_writer = DictWriter(f, fieldnames=dict_in.keys())
        if not append:
            dict_writer.writeheader()
        if not only_titles:
            dict_writer.writerow(dict_in)
        f.close()

def set_temperature():
    """
    initialize yocto temperature sensor
    :return: current temperature
    """
    # object tenp sens
    errmsg = YRefParam()
    if YAPI.RegisterHub("usb", errmsg) != YAPI.SUCCESS:
        sys.exit("init error :" + errmsg.value)
    cur_temp = YTemperature.FirstTemperature()
    print("Temp :  " + "%2.3f" % cur_temp.get_currentValue() + "°C (Ctrl-C to stop)")
    return cur_temp


def get_temperature():
    """
    initialize yocto temperature sensor
    :return: current temperature
    """
    # object tenp sens
    errmsg = YRefParam()
    if YAPI.RegisterHub("usb", errmsg) != YAPI.SUCCESS:
        sys.exit("init error :" + errmsg.value)
    cur_temp = YTemperature.FirstTemperature()
    return cur_temp.get_currentValue()


def print_exception(exception_details, printing_lock=None):
    """
    prints an exception to screen and to log
    :type exception_details: type(sys.exc_info())
    :param exception_details: desired printed exception
    :type printing_lock: threading.Lock()
    :param printing_lock: lock for printing
    """
    # print except - print_lock if None nothing else with
    exc_type, exc_obj, exc_trace = exception_details
    # deal with the exception
    if printing_lock is not None:
        with printing_lock:
            print('Exception details: ')
            print(exc_type)
            print(exc_obj)
            traceback.print_tb(exc_trace, limit=None, file=None)
    else:
        print('Exception details: ')
        print(exc_type)
        print(exc_obj)
        traceback.print_tb(exc_trace, limit=None, file=None)
    try:
        logging.warning('Exception details: exc_type = {}, exc_obj = {}'.format(exc_type, exc_obj))
    except Exception:
        pass


def encrypted_packet_decoder(processed_data):
    """
    extracts only the data relevant to the tester
    :type processed_data: dictionary (output of WiliotGateway.get_packets()))
    :return dictionary of relevant fields
    """
    packet_dict = {'advAddress': processed_data['adv_address'],
                   'packet_time': processed_data['time_from_start'],
                   'raw_data': processed_data['packet'],
                   'rssi': processed_data['rssi']}
    
    return packet_dict


def open_json(folder_path, file_path, default_values=None):
    """
    opens config json
    :type folder_path: string
    :param folder_path: the folder path which contains the desired file
    :type file_path: string
    :param file_path: the file path which contains the json
            (including the folder [file_path = folder_path+"json_file.json"])
    :type default_values: dictionary
    :param default_values: default values for the case of empty json
    :return: the desired json object
    """
    if not os.path.exists(folder_path):
        pathlib.Path(folder_path).mkdir(parents=True, exist_ok=True)
    
    file_exists = os.path.isfile(file_path)
    if not file_exists or os.stat(file_path).st_size == 0:
        # save the default values to json
        with open(file_path, "w") as out_file:
            json.dump(default_values, out_file)
        
        return json.load(open(file_path, "rb"))
    else:
        with open(file_path) as f:
            json_content = f.read()
        if len(json_content) == 0:
            with open(file_path, "w") as out_file:
                json.dump(default_values, out_file)
            json_content = json.load(open(file_path, "rb"))
        else:
            json_content = json.loads(json_content)
        return json_content


def open_json_cache(folder_path, file_path, default_values=None):
    """
    opens config json - for test/sgtin mode only
    :type folder_path: string
    :param folder_path: the folder path which contains the desired file
    :type file_path: string
    :param file_path: the file path which contains the json
            (including the folder [file_path = folder_path+"json_file.json"])
    :type default_values: dictionary
    :param default_values: default values for the case of empty json
    :return: the desired json object
    """
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)
    
    file_exists = os.path.isfile(file_path)
    if not file_exists:
        # save the default values to json
        with open(file_path, "w") as out_file:
            json.dump(default_values, out_file)
        with open(file_path, 'r') as f:
            json_content = json.load(f)
        return json_content
    else:
        with open(file_path, 'r') as f:
            json_content = json.load(f)

        for k, v in default_values.items():
            if k not in json_content.keys():
                json_content[k] = v
        with open(file_path, "w") as out_file:
            json.dump(json_content, out_file)
        return json_content


def printing_func(str_to_print, logging_entity, lock_print=None, do_log=True, logger_type='info', logger_name=None):
    """
    print and log
    :type str_to_print: string
    :param str_to_print: string to print
    :type logging_entity: string
    :param logging_entity: what thread is asking to log
    :type lock_print: threading.Lock
    :param lock_print: lock to prevent cutting printing in the middle
    :type do_log: bool
    :param do_log: if True log the data, do not log otherwise
    :type logger_type: string (info, debug, warning)
    :param logger_type: how to log the str_to_print
    """
    
    msg = str(logging_entity) + ':\t' + str(str_to_print)
    # if lock_print is not None:
    #     with lock_print:
    #         print(msg)
    # else:
    #     print(msg)
    # if do_log:
    #     if logger_type == 'info':
    #         logging.info(msg)
    #     elif logger_type == 'debug':
    #         logging.debug(msg)
    #     elif logger_type == 'warning':
    #         logging.warning(msg)
    if logger_name is None:
        if do_log:
            if lock_print is not None:
                with lock_print:
                    if logger_type == 'info':
                        logging.info(msg)
                    elif logger_type == 'debug':
                        logging.debug(msg)
                    elif logger_type == 'warning':
                        logging.warning(msg)
            else:
                if logger_type == 'info':
                    logging.info(msg)
                elif logger_type == 'debug':
                    logging.debug(msg)
                elif logger_type == 'warning':
                    logging.warning(msg)
        
        else:
            if lock_print is not None:
                with lock_print:
                    print(msg)
            else:
                print(msg)
    else:
        my_logger = logging.getLogger(logger_name)
        if do_log:
            if lock_print is not None:
                with lock_print:
                    if logger_type == 'info':
                        my_logger.info(msg)
                    elif logger_type == 'debug':
                        my_logger.debug(msg)
                    elif logger_type == 'warning':
                        my_logger.warning(msg)
            else:
                if logger_type == 'info':
                    my_logger.info(msg)
                elif logger_type == 'debug':
                    my_logger.debug(msg)
                elif logger_type == 'warning':
                    my_logger.warning(msg)
        
        else:
            if lock_print is not None:
                with lock_print:
                    print(msg)
            else:
                print(msg)


###########################################
#              Tester Classes             #
###########################################


class CounterThread(threading.Thread):
    """
    a class that manages and interacts with arduino_counter code

    Parameters:
        @type stop: Event
        @param stop: when set, close the port and exit
        @type printing_lock: threading.Lock
        @param printing_lock: a lock to access screen printing
        @type arduino_ports: list
        @param arduino_ports: list of arduino ports connected to PC

        global param tested: will use a global variable called tested

    Exceptions: None

    Events:
        listen/ waits on:
            stop => when set, close the port and exit

    Logging:
        debug: 'Warning: Could not decode counter data'
    """
    
    def __init__(self, stop: Event, printing_lock, arduino_ports):
        super(CounterThread, self).__init__()
        self.stop = stop
        self.printing_lock = printing_lock
        self.port = arduino_ports
        if len(self.port) > 1:
            print("too many USB Serial Device connected, can't figure out which is the counter")
        elif len(self.port) < 1:
            print("No USB Serial Device connected, please connect the counter")
        else:  # there is only one port of usb that is not GW
            self.port = self.port[0]
        self.baud = '9600'
        self.comPortObj = serial.Serial(self.port, self.baud, timeout=0.1)
    
    def run(self):
        global tested
        while not self.stop.isSet():
            data = self.comPortObj.readline()
            buf = b''
            if data.__len__() > 0:
                buf += data
                if b'\n' in buf:
                    try:
                        tmp = buf.decode().strip(' \t\n\r')
                        if "pulses detected" in tmp:
                            tested += 1
                    
                    except Exception:
                        with self.printing_lock:
                            print('Warning: Could not decode counter data')
                        logging.debug('Warning: Could not decode counter data')
                        continue
        
        self.comPortObj.close()


class TagsHandling:
    """
    controls the tags adv_address seen in the run, filtering bad packets, duplications etc.
    """
    
    def __init__(self, tags_list_len, lock_print=None, rssi_threshold=None, logging_thread=None,
                 only_add_tags_after_location_ends=False, add_to_black_list_after_locations=2):
        """
        :type tags_list_len: int
        :param tags_list_len: how long back to look for duplications
        :type lock_print: threading.lock
        :param lock_print: printing lock
        :type rssi_threshold: int
        :param rssi_threshold: max rssi value allowed for a packet
        :type logging_thread: string
        :param logging_thread: name of thread that will log the data
        :type only_add_tags_after_location_ends: bool
        :param only_add_tags_after_location_ends: if True will add the adv_addresses to
                                                  self.adv_addresses_in_current_location when calling to
                                                  self.set_new_location()
        :type add_to_black_list_after_locations: int (minimum is 2)
        :param add_to_black_list_after_locations: amount of locations a tag should appear in until
                                                  it will be added to black list
        """
        self.lock_print = lock_print
        self.rssi_threshold = int(rssi_threshold)
        self.only_add_tags_after_location_ends = only_add_tags_after_location_ends
        self.tags_list = []
        self.tags_list_len = tags_list_len
        self.cur_tag_adv_addr = None
        self.cur_tag_location = -1  # to make the first location be 0 (when calling set_new_location())
        self.cur_tag_min_rssi = np.inf  # big value, any real value will be much smaller
        self.group_id_hist = {}  # histogram of the group_id in this run
        self.black_list = []  # tags to ignore during the run
        self.black_list_candidates_hist = {}  # holds tags that will be added to black list if will reach to threshold
        self.add_to_black_list_after_locations = add_to_black_list_after_locations
        self.packets_in_current_location = []  # list of dictionaries (raw_data) of packets in this location
        self.adv_addresses_in_current_location = []  # list of adv_address from this location
        self.logging_thread = logging_thread
        self.problem_in_locations_hist = {'tag from black list': 0, 'bad group id': 0, 'above rssi threshold': 0,
                                          'packet from bridge': 0, 'duplication': 0, 'no singularization': 0}
        self.prev_location_problem_in_locations_hist = copy.copy(self.problem_in_locations_hist)
    
    def encrypted_packet_filter(self, raw_data, is_test_mode_packet=False):
        """
        checks if the packet is good in terms of RSSI, tag UID
        :type raw_data: dictionary (return value from gw_api.packet_listener (PROCESSED))
        :param raw_data: raw data of the tags
        :type is_test_mode_packet: bool
        :param is_test_mode_packet: is it a test mode packet (will only use the filtering by rssi)
        :return: is_good_packet - is the packet relevant for this run,
                 need_to_check_TBP - you should open wide window for catching more packets (using get_rates())
                                     due to Duplication/ non singularity issue (TBP =  time between packets)
                 reason_to_fail - string with reason for failing the packet (for logging in debug mode)
        """
        try:
            time.sleep(0)
            if is_test_mode_packet:
                if self.rssi_threshold is not None:
                    if int(raw_data['rssi']) > self.rssi_threshold:
                        msg = str(raw_data['packet']) + " - Packet rssi is too high"
                        printing_func(msg, self.logging_thread, self.lock_print, logger_type='debug')
                        reason_to_fail = 'rssi'
                        self.add_hist_val('above rssi threshold')
                        return False, False, reason_to_fail
                reason_to_fail = 'good'
                return True, False, reason_to_fail
            # check if the tag is in the black list
            if raw_data['adv_address'] in self.black_list:
                msg = str(raw_data['packet']) + " - tag that is in the black list (will be ignored)"
                printing_func(msg, self.logging_thread, self.lock_print, logger_type='debug')
                reason_to_fail = 'blackList'
                self.add_hist_val('tag from black list')
                return False, False, reason_to_fail
            
            else:
                self.packets_in_current_location.append(raw_data)
                
                # check if it is a bridge packet
                if self.is_packet_from_bridge(raw_data):
                    self.add_tag_to_black_list(raw_data['adv_address'])
                    msg = str(raw_data['packet']) + " - packet generated by Wiliot bridge found, " \
                                                    "will add it to the black list"
                    printing_func(msg, self.logging_thread, self.lock_print, logger_type='debug')
                    reason_to_fail = 'bridge'
                    self.add_hist_val('packet from bridge')
                    return False, False, reason_to_fail
                
                # check if the tag group_id is the same as the rest of the tags
                bad_group_id = False
                if raw_data['group_id'] is not None:
                    if raw_data['group_id'] not in self.group_id_hist.keys():
                        if len(self.group_id_hist) > 0:
                            max_group_id = max(self.group_id_hist.values())
                            for i, (group, group_val) in enumerate(self.group_id_hist.items()):
                                if group_val == max_group_id and not raw_data['group_id'] == group:
                                    bad_group_id = True
                        self.group_id_hist[raw_data['group_id']] = 1
                    else:
                        if len(self.group_id_hist) > 0:
                            max_group_id = max(self.group_id_hist.values())
                            for i, (group, group_val) in enumerate(self.group_id_hist.items()):
                                if group_val == max_group_id and not raw_data['group_id'] == group:
                                    bad_group_id = True
                        self.group_id_hist[raw_data['group_id']] += 1
                    if bad_group_id:
                        msg = str(raw_data['packet']) + " - packet with wrong group_id found (group_id = " + \
                              str(raw_data['group_id']) + ', group_id seen so far (from all packets) = ' + \
                              str(self.group_id_hist) + ')'
                        printing_func(msg, self.logging_thread, self.lock_print, logger_type='debug')
                        reason_to_fail = 'group_id'
                        self.add_hist_val('bad group id')
                        return False, False, reason_to_fail
                
                # check if the RSSI is good
                if self.rssi_threshold is not None:
                    if int(raw_data['rssi']) > self.rssi_threshold:
                        msg = str(raw_data['packet']) + " - Packet rssi is too high"
                        printing_func(msg, self.logging_thread, self.lock_print, logger_type='debug')
                        reason_to_fail = 'rssi'
                        self.add_hist_val('above rssi threshold')
                        return False, False, reason_to_fail
                
                is_new_tag = self.add_new_tag_to_tags_list(raw_data)
                
                # check if the tag was already caught in the GW
                if not is_new_tag:
                    msg = str(raw_data['packet']) + " - Duplication from a tag we have seen before (advAddress = " \
                          + raw_data['adv_address'] + ")"
                    printing_func(msg, self.logging_thread, self.lock_print, logger_type='debug')
                    reason_to_fail = 'duplication'
                    self.add_hist_val('duplication')
                    return True, True, reason_to_fail
                
                # check if this packet is from new tag
                if is_new_tag and self.cur_tag_adv_addr != raw_data['adv_address'] and \
                        self.cur_tag_adv_addr is not None:
                    msg = str(raw_data['packet']) + " - no singularity issue (did not see this tag before), " \
                                                    "new tag advAddress = " + str(raw_data['adv_address']) \
                          + ", new tag rssi = " + str(raw_data['rssi']) \
                          + ', current tag rssi = ' + str(self.cur_tag_min_rssi)
                    printing_func(msg, self.logging_thread, self.lock_print, logger_type='debug')
                    reason_to_fail = 'no singularity'
                    self.add_hist_val('no singularization')
                    return True, True, reason_to_fail
                
                self.cur_tag_adv_addr = raw_data['adv_address']
                if self.cur_tag_min_rssi > int(raw_data['rssi']):
                    self.cur_tag_min_rssi = int(raw_data['rssi'])
                
                reason_to_fail = 'good'
                return True, False, reason_to_fail
        
        except Exception as e:
            printing_func('Exception during encrypted_packet_filter: ' + str(e), self.logging_thread,
                          self.lock_print, logger_type='debug')
            return False, False, 'exception'
    
    def is_packet_from_bridge(self, raw_data):
        """
        check if the packet generated by Wiliot bridge
        :type raw_data: dictionary (return value from gw_api.packet_listener)
        :param raw_data: raw data of the tags
        :return: True if the tag is bridge packet, False otherwise
        """
        chars_starts_with10_bits = ['8', '9', 'A', 'B', 'C', 'D', 'E', 'F']
        if str(raw_data['packet'])[0].upper() in chars_starts_with10_bits and 'AFFD' in str(raw_data['packet']):
            return True
    
    def add_new_tag_to_tags_list(self, raw_data):
        """
        adds a new tag's adv_address to the list of the tags we have seen before
        :type raw_data: dictionary (return value from gw_api.packet_listener)
        :param raw_data: raw data of the tags
        :return: True if the tag is a new tag, False otherwise
        """
        if self.only_add_tags_after_location_ends:
            if raw_data['adv_address'] not in self.adv_addresses_in_current_location:
                self.adv_addresses_in_current_location.append(raw_data['adv_address'])
            if raw_data['adv_address'] not in self.tags_list:
                return True
            else:
                return False
        
        elif raw_data['adv_address'] not in self.tags_list:
            if len(self.tags_list) <= self.tags_list_len:
                self.tags_list.append(str(raw_data['adv_address']))
            else:
                del self.tags_list[0]
                self.tags_list.append(str(raw_data['adv_address']))
            return True
        else:
            return False
    
    def add_tag_to_black_list(self, adv_address):
        """
        adds a tag's adv_address to the black list (tags to ignore during the run)
        :type adv_address: string
        :param adv_address: adv_address of the tag to add to black list
        :return: True if the tag is a new tag, False otherwise
        """
        if adv_address not in self.black_list:
            self.black_list.append(adv_address)
            return True
        else:
            return False
    
    def add_hist_val(self, key):
        """
        :type key: string
        :param key: one of the keys in self.problem_in_locations_hist
        """
        if key not in self.prev_location_problem_in_locations_hist.keys() or \
                key not in self.problem_in_locations_hist.keys():
            return
        
        if self.prev_location_problem_in_locations_hist[key] == self.problem_in_locations_hist[key]:
            self.problem_in_locations_hist[key] += 1
    
    def set_new_location(self):
        """
        reset all packet rates from previous tags, will start rates comparison with tags received from now
        until this function will be called again
        """
        for adv_address in self.adv_addresses_in_current_location:
            if adv_address not in self.tags_list:
                if self.only_add_tags_after_location_ends:
                    if len(self.tags_list) <= self.tags_list_len:
                        self.tags_list.append(str(adv_address))
                    else:
                        del self.tags_list[0]
                        self.tags_list.append(str(adv_address))
            else:
                if adv_address not in self.black_list_candidates_hist.keys():
                    self.black_list_candidates_hist[adv_address] = 2  # already appeared in previous location
                else:
                    self.black_list_candidates_hist[adv_address] += 1
                if self.black_list_candidates_hist[adv_address] >= self.add_to_black_list_after_locations:
                    self.add_tag_to_black_list(adv_address)
        
        self.adv_addresses_in_current_location = []
        self.packets_in_current_location = []
        self.cur_tag_location += 1
        self.cur_tag_adv_addr = None
        self.prev_location_problem_in_locations_hist = copy.copy(self.problem_in_locations_hist)
    
    def is_tag_appeared_before(self, adv_addr):
        return adv_addr in self.tags_list
    
    def get_black_list_size(self):
        return len(self.black_list)


###########################################
#              Tester GUI             #
###########################################


def credentials_gui():
    """
    open GUI for getting user_name, password and owner_id from user
    """
    layout = [
        [SimGUI.Text('Please insert FusionAuth Credentials for uploading and serialization this run data')],
        [SimGUI.Text('User name (email address):'),
         SimGUI.InputText('', key='user_name')],
        [SimGUI.Text('Password:'),
         SimGUI.InputText('', key='password')],
        [SimGUI.Text('Owner Id:'),
         SimGUI.InputText('wiliot-ops', key='owner_id')],
        [SimGUI.Submit()]]
    
    window = SimGUI.Window('User Credentials', layout)
    while True:
        event, values = window.read()
        if event == 'Submit':
            break
        elif event is None:
            print('User exited the program')
            window.close()
            exit()
    
    window.close()
    return values


def upload_conclusion(failed_tags=None, succeeded_csv_uploads=None):
    """
    :type failed_tags: list or None
    :param failed_tags: list of dictionaries ({'tag_id': XXX, 'status': YYY}) of tags that failed in serialization
    :type succeeded_csv_uploads: bool or None
    :param succeeded_csv_uploads: if True - upload of csv has succeeded
    """
    if failed_tags is None or len(failed_tags) == 0:
        serialization_line = [SimGUI.Text("Serialization succeeded", border_width=10, background_color='green')]
    else:
        serialization_line = [
            SimGUI.Text("Serialization failed for few tags, please check status in log file", border_width=10,
                        background_color='orange')]
    
    if succeeded_csv_uploads is None or succeeded_csv_uploads:
        csv_uploads_line = [SimGUI.Text("Upload CSV's succeeded", border_width=10, background_color='green')]
    else:
        csv_uploads_line = [SimGUI.Text("Upload CSV's Failed, look in log file for more details", border_width=10,
                                        background_color='orange')]
    
    if failed_tags is None:
        layout = [csv_uploads_line, [SimGUI.Exit()]]
    else:
        layout = [serialization_line, csv_uploads_line, [SimGUI.Exit()]]
    
    window2 = SimGUI.Window('Conclusion', layout, keep_on_top=True)
    event, values = window2.read()
    window2.close()


###########################################
#              Cloud Functions            #
###########################################


def get_cloud_error_message(status_code):
    if status_code is None:
        message = 'upload failed with unknown error code, please look inside log and reach wiliot'
    elif status_code == 500:
        message = 'There was a problem at the Server side (status 500), ' \
                  'please try to upload again or reach Wiliot'
    
    elif status_code == 401:
        message = 'There was a problem with the authorization (status 401),\n ' \
                  'Please make sure your user name and password in wiliot_testers/' \
                  'offline/configs/user_configs.json is correct or reach Wiliot'
    
    elif str(status_code)[0:2] == '40':
        message = 'There was a problem with the request (status 40*), ' \
                  'please try to upload again or reach Wiliot \n' \
                  'Status code is: ' + str(status_code)
    elif status_code == 200:
        message = 'Task ended successfully'
    else:
        message = 'There was a problem, Status code is: ' + str(status_code)
    
    return message


def check_serialization_exception_queues(serialization_threads_working, end_the_fails=False, printing_lock=None,
                                         to_logging=False):
    """
    checks if there is any response for any serialization thread, if it is good (success) will terminate the thread,
    else will raise exception
    :type serialization_threads_working: list of SerializationAPI
    :param serialization_threads_working: the Serialization process that needs to be monitored
    :type end_the_fails: bool
    :param end_the_fails: if True, kill all threads that had an exception
    :type printing_lock: threading.Lock
    :param printing_lock: printing lock
    :type to_logging: bool
    :param to_logging: if true, write to logging.debug the result of the upload
    """
    serialize_status = True
    for thread in serialization_threads_working:
        tags_in_batch = []
        for i in range(len(thread.batch_dictionary['upload_data'])):
            tags_in_batch.append(thread.batch_dictionary['upload_data'][i]['tagId'])
        if not thread.exception_queue.empty():
            exception_details = thread.exception_queue.get()
            # using logging.warning that will be parsed to errors
            # print_exception(exception_details, printing_lock=printing_lock)
            logging.warning(str(exception_details))
            exc_type, exc_obj, exc_trace = exception_details
            if end_the_fails:
                thread.kill_thread.set()
                thread.join(timeout=0.5)
            logging.warning('Serialization failed for tags ' + str(tags_in_batch))
            serialize_status = False
    return serialize_status


def check_serialization_response(serialization_threads_working, printing_lock=None, to_logging=False):
    """
    checks if there is any response for any serialization thread, if it is good (success) will terminate the thread,
    else will raise exception
    :type serialization_threads_working: list of SerializationAPI
    :param serialization_threads_working: the Serialization process that needs to be monitored
    :type printing_lock: threading.Lock
    :param printing_lock: printing lock
    :type to_logging: bool
    :param to_logging: if true, write to logging.debug the result of the upload
    """
    for thread in serialization_threads_working:
        response = thread.get_response()
        if 'uploaded Successfully' in response:
            thread.kill_thread.set()
            thread.join(timeout=0.5)
        else:
            pass


def close_all_serialization_processes_when_they_done(serialization_threads_working, try_serialize_again=None,
                                                     printing_lock=None, to_logging=False):
    """
    terminates all serialization threads when they are done
    :type serialization_threads_working: list of SerializationAPI
    :param serialization_threads_working: the Serialization process that needs to be monitored
    :type try_serialize_again: threading.Event
    :param try_serialize_again: if exists it will be set to set and it will make all
                                threads that failed for the first time to try again
    :type printing_lock: threading.Lock
    :param printing_lock: printing lock
    :type to_logging: bool
    :param to_logging: if true, write to logging.debug the result of the upload
    :return: list of dictionaries ({'tag_id': XXX, 'status': YYY}) of tags which have failed to serialize
    """
    if len(serialization_threads_working) > 0:
        message = 'please wait until all serialization processes end,\n' \
                  'will let you know when it is done (not more than a minute)'
        if printing_lock is not None:
            printing_func(str_to_print=message, logging_entity="serializationThread",
                          lock_print=printing_lock, do_log=to_logging)
        else:
            print(message)
    if try_serialize_again is not None:
        try_serialize_again.set()
    
    failed_tags = []
    # tmp_serialization_threads_working = copy.deepcopy(serialization_threads_working)
    while len(serialization_threads_working) > 0:
        locs_to_pop = []
        for loc, thread in enumerate(serialization_threads_working):
            thread.kill_thread.set()
            response = thread.get_response()
            tags_in_batch = []
            for i in range(len(thread.batch_dictionary['upload_data'])):
                tags_in_batch.append(thread.batch_dictionary['upload_data'][i]['tagId'])
            
            if thread.is_done:
                locs_to_pop.append(loc)
                failed_tags += thread.failed_tags
        
        for loc in reversed(locs_to_pop):
            serialization_threads_working[loc].kill_thread.set()
            serialization_threads_working[loc].join(timeout=0.5)
            serialization_threads_working.pop(loc)
    
    for thread in serialization_threads_working:
        thread.kill_thread.set()
        thread.join(timeout=0.5)
    
    time.sleep(0.5)
    
    # if len(serialization_threads_working) > 0:
    message = '\nAll serialization processes ended\n'
    if len(failed_tags) > 0:
        message += '\n-------------------------------------\n'
        message += 'failed tags in serialization process:\n'
        for i in range(len(failed_tags)):
            message += str(failed_tags[i]) + '\n'
        message += '-------------------------------------\n'
    else:
        message += 'serialization of all tags in files succeeded'
    
    if printing_lock is not None:
        printing_func(str_to_print=message, logging_entity="serializationThread",
                      lock_print=printing_lock, do_log=to_logging)
    else:
        print(message)
    
    return failed_tags


###########################################
#         Multiple Events Functions       #
###########################################


def or_set(self):
    """
    :param self: threading.Event()
    """
    self._set()
    self.changed()


def or_clear(self):
    """
    :param self: threading.Event()
    """
    self._clear()
    self.changed()


def orify(e, changed_callback):
    """
    combines the events
    :type e: threading.Event()
    :param e: the event to be combined into
    :type changed_callback: threading.Event()
    :param changed_callback: the event to be combined into
    """
    e._set = e.set
    e._clear = e.clear
    e.changed = changed_callback
    e.set = lambda: or_set(e)
    e.clear = lambda: or_clear(e)


def or_event_set(*events):
    """
    creates an event the combines multiple events
    :type threading.Event() list
    :return: the said event
    """
    or_event = threading.Event()
    
    def changed():
        """
        makes sure all of the events are clear when the orEvent is created
        """
        bools = [e.is_set() for e in events]
        if any(bools):
            or_event.set()
        else:
            or_event.clear()
    
    for e in events:
        orify(e, changed)
    changed()
    return or_event


###########################################
#               Post-Process              #
###########################################


def process_encrypted_tags_data(data, packet_threshold, tester_type=TesterName.OFFLINE, fail_this_tag=False,
                                is_debug_mode=False, packets_time_diff=None, adv_of_selected_tag=None):
    """
    :type data: list
    :param data: list of dict with tag_data
    :type packet_threshold: int
    :param packet_threshold: amount of packets for tag to pass
    :type tester_type: TesterName()
    :param tester_type: what tester uses this function
    :type fail_this_tag: bool
    :param fail_this_tag: indicates if 'status' field should be 'Failed' (in case of duplications)
    :type is_debug_mode: bool
    :param is_debug_mode: if True adds more columns to tags_data, if False will not add anything
                          (only in offline tester)
    :type packets_time_diff: dictionary
    :param packets_time_diff: packets time differences
    :type adv_of_selected_tag: string
    :param adv_of_selected_tag: add_address of the selected tag
    :rtype: dict
    :rvalue: dictionary with defined values
    """
    df = pd.DataFrame(data)
    adv_addr = df['advAddress'].iloc[0]
    tag_location = df['tagLocation'].iloc[0]
    common_run_name = df['commonRunName'].iloc[0]  # get first value (common value to all tags in run)
    received_packet_count = len(data)  # data is list of dicts where every dict is from one packet
    raw_data = df[['packet_time', 'raw_data']].to_json(orient="records")
    
    status = 'Failed'
    if received_packet_count >= packet_threshold and not fail_this_tag:
        status = 'Passed'
    
    if tester_type.value == 'offline':
        external_id = set_external_id(df['externalId'].iloc[0])
        ret_dict = {'advAddress': adv_addr,
                    'tagLocation': tag_location,
                    'externalId': external_id,
                    'status': status,
                    'commonRunName': common_run_name}
        
        if 'temperatureFromSensor' in df.columns:  # exists only if temperature sensor is enabled
            ret_dict['temperatureFromSensor'] = df['temperatureFromSensor'].iloc[0]
        good_df = df[df['packetStatus'].isin(['good', 'no singularity'])]
        try:
            good_df = good_df[good_df['advAddress'] == adv_of_selected_tag]
            ret_dict['Ttfgp'] = min(good_df['packet_time'])
        except Exception:
            ret_dict['Ttfgp'] = None
        if is_debug_mode:
            if 'temperatureFromSensor' in df.columns:  # exists only if temperature sensor is enabled
                ret_dict['temperatureFromSensor'] = df['temperatureFromSensor'].iloc[0]
            ret_dict['AdvAddressesInLocation'] = list(df['advAddress'].unique())
            if adv_of_selected_tag is not None:
                ret_dict['advAddress'] = adv_of_selected_tag
            raw_data = df[['packetStatus', 'packet_time', 'raw_data']].to_json(orient="records")
            if packets_time_diff is not None:
                ret_dict['packetsTimeDiff'] = packets_time_diff
        
        ret_dict['rawData'] = raw_data
        return ret_dict
    else:
        return {'advAddress': adv_addr,
                'tagLocation': tag_location,
                'status': status,
                'commonRunName': common_run_name,
                'rawData': raw_data}


def set_external_id(external_id=None):
    """
    returns external ID value to calling function (returns 'None' for unconverted tags)
    @type external_id: str (default is None)
    @param external_id: ID to print on tag
    @rtype: str
    @rvalue: 'None' (for unconverted) or external_id string
    """
    if external_id is not None:
        return external_id
    else:
        return 'None'


###########################################
#                   Log                   #
###########################################


def snake_to_camel(word):
    ret_val = ''.join(x.capitalize() or '_' for x in word.split('_'))
    ret_val = ret_val[:1].lower() + ret_val[1:] if ret_val else ''
    return ret_val


def collect_errors(log_path='', log_file_lines=''):
    """
    collect errors from logfile
    @type log_path: str
    @param log_path: log path file
    @type log_file_lines: str
    @param log_file_lines: if empty (''), read log file
    @rtype: list
    @rvalue: list of errors occurred in test
    """
    reel_data = {'tested': 0, 'passed': 0, 'includingUnderThresholdPassed': 0}
    errors = []
    if log_file_lines == '':
        log_file = open(file=log_path)
        log_file_lines = log_file.readlines()
    tag_data = {'tagLocation': None, 'externalId': None, 'status': None, 'packets': None}
    for line in log_file_lines:
        if 'User set up is:' in line:
            parts = [p for p in line.split("{")]
            # the dictionary from main window
            parts2 = [p for p in parts[1].split("}")]
            config_dict_from_user = ast.literal_eval('{' + parts2[0] + '}')
            # the dictionary from memory
            parts2 = [p for p in parts[2].split("}")]
            config_dict_from_memory = ast.literal_eval('{' + parts2[0] + '}')
            # the printing string dictionary
            try:
                parts3 = [p for p in parts[3].split("}")]
                printing_dict_from_memory = ast.literal_eval('{' + parts3[0] + '}')
            except Exception:
                printing_dict_from_memory = {}
            
            config_dict = {**config_dict_from_user, **config_dict_from_memory}
            config_dict = {**config_dict, **printing_dict_from_memory}
            for key, value in config_dict.items():
                reel_data[key] = value
            reel_data['tested'] = 0
        line = line.strip()
        if 'DEBUG' in line:
            if 'Duplication' in line:
                tmp = {'errorLocation': reel_data['tested'], 'errorString': None}
                parts = [p for p in line.split("DEBUG ")]
                tmp['errorString'] = parts[1]
                errors.append(tmp)
            
            if 'Packet rssi is too high' in line:
                tmp = {'errorLocation': reel_data['tested'], 'errorString': None}
                parts = [p for p in line.split("DEBUG ")]
                tmp['errorString'] = parts[1]
                errors.append(tmp)
            
            if 'Received packet that is too short' in line:
                tmp = {'errorLocation': reel_data['tested'], 'errorString': None}
                parts = [p for p in line.split("DEBUG ")]
                tmp['errorString'] = parts[1]
                errors.append(tmp)
            
            if 'Exception happened:, exc_type = ' in line:
                tmp = {'errorLocation': reel_data['tested'], 'errorString': None}
                parts = [p for p in line.split("DEBUG ")]
                tmp['errorString'] = parts[1]
                errors.append(tmp)
        
        if 'INFO' in line:
            parts = [p for p in line.split("INFO")]
            
            if 'Tag location is: ' in parts[1]:
                parts[1] = parts[1].strip('Tag location is: ')
                parts = [p for p in parts[1].split(',')]
                tag_data['tagLocation'] = int(parts[0])
                reel_data['tested'] = int(tag_data['tagLocation']) + 1  # locations starts from 0
        
        if 'WARNING ' in line:  # the space is here to ignore the 'Warnings' print from the printer
            tmp = {'errorLocation': reel_data['tested'], 'errorString': None}
            parts = [p for p in line.split("WARNING ")]
            tmp['errorString'] = parts[1]
            errors.append(tmp)
    
    if len(errors) != 0:
        return errors
    return None


def update_json_field(path, key, value):
    """
    update json config file value
    :type path: string
    :param path: json config file path
    :type key: string
    :param key: key to change in json file
    :type value: string
    :param value: value to change @key to
    :return:
    """
    with open(path, "r+") as jsonFile:
        config_data = json.load(jsonFile)
        
        config_data[key] = value
        
        jsonFile.seek(0)
        json.dump(config_data, jsonFile)
        jsonFile.truncate()


class CsvLog:
    """
    Class for csv logs
    """
    
    def __init__(self, header_type, path, headers=None, tester_type=TesterName.NONE, temperature_sensor=False,
                 is_debug_mode=False):
        """
        :type header_type: HeaderType
        :param header_type:
        :param path:
        :param headers:
        :param tester_type:
        :type temperature_sensor: bool
        :param temperature_sensor: if True -> temperature sensor enabled -> add header. else: sensor is disabled
        :type is_debug_mode: bool
        :param is_debug_mode: if True adds more columns to tags_data, if False will not add anything
                              (only in offline tester)
        """
        self.tags_header = ['advAddress', 'tagLocation', 'status', 'commonRunName', 'rawData']
        
        self.run_header = ['testerStationName', 'commonRunName', 'batchName', 'testerType',
                           'comments', 'errors',
                           'timeProfile', 'txPower', 'energizingPattern',
                           'tested', 'passed', 'yield', 'inlay']
        
        self.packets_header = ['commonRunName', 'encryptedPacket', 'time']
        
        if header_type.value == HeaderType.TAG.value:
            self.header = self.tags_header
        elif header_type.value == HeaderType.RUN.value:
            self.header = self.run_header
        elif header_type.value == HeaderType.PACKETS.value:
            self.header = self.packets_header
        else:
            self.header = []
        self.path = path
        
        # add additional columns per tester:
        self.temperature_sensor_enable = temperature_sensor
        self.add_headers(headers, header_type, tester_type, is_debug_mode)
    
    def add_headers(self, headers=None, header_type=HeaderType.RUN, tester_type=TesterName.NONE, is_debug_mode=False):
        """
        added additional headers to output file according to file type and tester type
        :type headers:
        :param headers: additional headers for output file (determined by user)
        :type header_type: HeaderType
        :param header_type: determines if the file generated stores run data or tags data
        :type tester_type: TesterName
        :param tester_type: determines unique headers to add according to tester type
        :type is_debug_mode: bool
        :param is_debug_mode: if True adds more columns to tags_data, if False will not add anything
                              (only in offline tester)
        """
        if headers is None:
            if header_type.value == HeaderType.RUN.value:
                if tester_type.value == TesterName.OFFLINE.value:
                    added_headers = ['testTime', 'maxTtfp', 'converted', 'surface', 'tagGen',
                                     'toPrint', 'passJobName', 'printingFormat', 'stringBeforeCounter',
                                     'digitsInCounter', 'gwVersion', 'desiredPass',
                                     'desiredTags', 'firstPrintingValue', 'missingLabel', 'maxMissingLabels']
                    # TODO - add this list instead af the previous when cloud supports it
                    # added_headers = ['testTime', 'maxTtfp', 'converted', 'surface', 'tagGen',
                    #                  'toPrint', 'passJobName', 'printingFormat', 'stringBeforeCounter',
                    #                  'digitsInCounter', 'gwVersion', 'secondEnergizingPattern', 'desiredPass',
                    #                  'desiredTags', 'firstPrintingValue', 'missingLabel', 'maxMissingLabels',
                    #                  'temperatureSensorEnable', 'wiliotPackageVersion']
                
                elif tester_type.value == TesterName.TAL15K.value:
                    added_headers = ['testersGwVersion', 'chargerGwVersion', 'rowsNum', 'columnsNum',
                                     'numOfTesters', 'chargingTime', 'timePerTag']
                elif tester_type.value == TesterName.CONVERSION.value:
                    added_headers = []
                elif tester_type.value == TesterName.SAMPLE.value:
                    added_headers = ['responded', 'responding[%]', 'passed[%]', 'validTbp[%]', 'testStatus',
                                     'operator', 'testTime', 'runStartTime', 'runEndTime', 'antennaType',
                                     'surface', 'numChambers', 'gwVersion', 'pyWiliotVersion',
                                     'bleAttenuation', 'loraAttenuation', 'testTimeProfilePeriod',
                                     'testTimeProfileOnTime', 'ttfpAvg', 'tbpAvg', 'tbpStd', 'rssiAvg', 'maxTtfp',
                                     'controlLimits', 'hwVersion', 'sub1gFrequency', 'failBinStr', 'failBin',
                                     'uploadToCloud']
                else:
                    added_headers = []
            elif header_type.value == HeaderType.TAG.value:
                if tester_type.value == TesterName.OFFLINE.value:
                    added_headers = ['externalId']
                    if self.temperature_sensor_enable:
                        added_headers.append('temperatureFromSensor')
                    if is_debug_mode:
                        added_headers.append('AdvAddressesInLocation')
                        added_headers.append('packetsTimeDiff')
                        added_headers.append('Ttfgp')
                elif tester_type.value == TesterName.SAMPLE.value:
                    added_headers = ['externalId', 'temperature_from_sensor']
                else:
                    added_headers = []
            elif header_type.value == HeaderType.PACKETS.value:
                added_headers = []
                if tester_type.value == TesterName.OFFLINE.value:
                    added_headers = ['externalId', 'tagLocation', 'packetStatus', 'chosenTagInLocation',
                                     'chosenTagStatus', 'rssi', 'advAddr', 'gw_tx_power', 'attenuation', 'gw_time',
                                     'packet_data', 'tbp', 'test_num', 'test_duration']
                    if self.temperature_sensor_enable:
                        added_headers.append('temperatureFromSensor')
                elif tester_type.value == TesterName.SAMPLE.value:
                    added_headers = ['externalId', 'chamber', 'status',
                                     'state(tbp_exists:0,no_tbp:-1,no_ttfp:-2,dup_adv_address:-3)', 'fail_bin',
                                     'reel', 'ttfp', 'tbp', 'adv_address', 'temperature_from_sensor']
            else:
                added_headers = []
        else:
            added_headers = headers
        self.header += added_headers
    
    def open_csv(self):
        """
        create csv file if not exist
        """
        # dir_path = self.path[:(len(self.path) - (len(self.path.split('/')[-1])+1))]
        dir_path = dirname(self.path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        if not os.path.isfile(self.path):
            with open(self.path, 'w', newline='') as write_obj:
                csv_writer = writer(write_obj)
                csv_writer.writerow(self.header)
    
    def append_dict_as_row(self, data_to_append):
        """
        append data list to existing csv
        @type data_to_append: list
        @param data_to_append: dictionary for each new row
        """
        # Open file in append mode
        with open(self.path, 'a+', newline='') as write_obj:
            # Create a dict writer object from csv module
            csv_writer = DictWriter(write_obj, fieldnames=self.header)
            # Add contents of list as last rows in the csv file
            csv_writer.writerows(data_to_append)
    
    def append_list_as_row(self, data_to_append):
        """
        append data list to existing csv
        @type data_to_append: list
        @param data_to_append: values to append to csv
        """
        # Open file in append mode
        with open(self.path, 'a+', newline='') as write_obj:
            # Create a writer object from csv module
            csv_writer = writer(write_obj)
            # Add contents of list as last row in the csv file
            csv_writer.writerow(data_to_append)
    
    def append_table(self, data_table):
        """
        append data table to existing csv (using append_list_as_row)
        @type: data_table: list (list of lists)
        @param data_table: values to append to csv
        """
        col = 0  # the table has only 1 column
        for i in range(len(data_table)):
            self.append_list_as_row(self.dict_to_list(data_table[i][col]))
    
    def dict_to_list(self, data):
        """
        convert dict to list
        @type data: dict
        @param data:
        @rtype: list
        """
        res_list = []
        for title in self.header:
            if title in data.keys():
                res_list.append(data[title])
            else:
                res_list.append('')
        return res_list
    
    def override_run_data(self, run_data, path_=None):
        """
        override run_data.csv at the end of run
        @type run_data: dict
        @param run_data: run configurations and results such as passed, tested, yield, etc.
        """
        if path_ is None:
            path_ = self.path
        
        with open(path_, 'w', newline='') as write_obj:
            csv_writer = writer(write_obj)
            csv_writer.writerow(self.header)
            # self.append_list_as_row(self.path, self.dict_to_list(run_data))
            csv_writer.writerow(self.dict_to_list(run_data))


class colors:
    # Printing colors by log level:
    MESSAGE = '\033[36m'
    
    DEBUG = '\033[36m'
    INFO = '\033[0;30;47m'
    WARNING = '\033[35m'
    ERROR = '\033[91m'
    CRITICAL = '\033[31m'
    
    END = '\033[0m'


class CustomFormatter(logging.Formatter):
    format = "%(message)s"
    module = "Module: %(module)s"
    line = 'File "%(pathname)s", line %(lineno)s\n'
    exception = '%(exc_info)s\n'
    #     moduleAndLine = module + "  " + line + " :\n"
    
    FORMATS = {
        logging.DEBUG: colors.DEBUG + format + colors.END,
        logging.INFO: colors.INFO + format + colors.END,
        logging.WARNING: colors.WARNING + format + colors.END,
        logging.ERROR: colors.ERROR + format + colors.END,
        logging.CRITICAL: colors.CRITICAL + format + colors.END
    }
    
    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def setLogger(loggerName, log_level, outputFile=None, file_mode='w+'):
    # Create the logger
    logger = logging.getLogger(loggerName)
    logger.setLevel(log_level)
    
    # Create stream handler
    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(CustomFormatter())
    streamHandler.setLevel(getattr(logging, log_level))
    
    # Create file handler
    if outputFile is not None:
        fileHandler = logging.FileHandler(outputFile, file_mode)
        fileHandler.setLevel(logging.DEBUG)
        logger.addHandler(fileHandler)
        formatter = logging.Formatter('%(asctime)s - %(message)s')
        fileHandler.setFormatter(formatter)
    
    # Add the stream and file handler to the logger
    logger.addHandler(streamHandler)
    return logger


def changeFileHandler(logger, newHandlerFile, file_mode='w+', append_handler=False):
    if not append_handler:
        for hdlr in logger.handlers[:]:  # remove all old handlers
            if 'filehandler' in str(hdlr).lower():
                logger.removeHandler(hdlr)
    fileHandler = logging.FileHandler(newHandlerFile, file_mode)
    fileHandler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(message)s')
    fileHandler.setFormatter(formatter)
    logger.addHandler(fileHandler)
    return logger


def removeFileHandler(logger, fileHandler):
    for hdlr in logger.handlers[:]:  # remove all old handlers
        if 'filehandler' in str(hdlr).lower() and hdlr.baseFilename == fileHandler:
            logger.removeHandler(hdlr)
    return logger


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """
    
    def __init__(self, logger, level):
        self.logger = logger
        self.level = level
        self.linebuf = ''
    
    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.level, line.rstrip())
    
    def flush(self):
        pass


def format_exception_details(exception_details):
    """
    Temporary till exceptions handling will be more robustic...
    Parameters
    ----------
    exception_details - as retured from sys.exc_info()

    Returns
    -------
    Exception string, can be forwarded into a printing function
    """
    return "Exception: {},\n Traceback: {}".format(exception_details[1], *traceback.format_tb(exception_details[2]))


class SerializationWorker(threading.Thread):
    def __init__(self, client, input_q, output_q, failed_tags_q, max_retrials=100, sample_timeout_dt=10,
                 min_delay_between_batches=0):
        super(SerializationWorker, self).__init__()
        if client is None:
            raise ValueError("client can't be none!")
        self.client = client
        self.input_q = input_q
        self.output_q = output_q
        self.failed_tags_q = failed_tags_q
        self._stop_event = threading.Event()
        self.max_retrials = max_retrials
        self.sample_timeout_dt = sample_timeout_dt
        self.min_delay_between_batches = min_delay_between_batches
    
    def stop(self):
        self._stop_event.set()
    
    def stopped(self):
        return self._stop_event.is_set()
    
    def run(self):
        print("Starting worker")
        while not self.stopped():
            try:
                curr_batch = self.input_q.get(timeout=self.sample_timeout_dt)
                print("Starting batch {}".format(curr_batch['upload_data']))
            except Empty:
                # print("Failed to get batch for worker\n")
                time.sleep(5)
                continue
            # TODO: add Signal for the worker to stop at the Q instead of timeout:
            # if curr_batch['stop']:
            #     self.stop()
            #     break
            
            # print(len(curr_batch['upload_data'][0]['payload']))
            if 'retrials' not in curr_batch:
                curr_batch['retrials'] = 1
            else:
                if curr_batch['retrials'] > self.max_retrials:
                    continue
                else:
                    curr_batch['retrials'] += 1
            # Try serialization:
            try:
                result = self.client.batch_serialize_tags(curr_batch['upload_data'])
                if isinstance(result, dict):
                    if len(result['data']) < len(curr_batch['upload_data']):
                        print('Error: There are {} tags which did not get response from the cloud! '
                              'Status code: {} \n  Reuested batch:\n {} \n Got response:\n {}'.format(
                            len(curr_batch['upload_data']) - len(result['data']), result['status_code'],
                            curr_batch['upload_data'], result['data']))
                success = True
            except WiliotCloudError as e:
                result = eval(str(e))
                success = False
            except Exception as e:
                raise (e)
            
            # repeat serialization only if error is not 200 or 207:
            if not success:
                print(f"Batch failed: {curr_batch} with response {result['status_code']}")
                if result['status_code'] == 207:
                    for tag_result in result['data']:
                        if not tag_result['isSuccess']:
                            # the same request was already executed
                            if "Tag ID was already assigned" in tag_result['message']:
                                tag_result['message'] = tag_result[
                                                            'message'] + '. Tag was already serialized with this tagID'
                            self.failed_tags_q.put(tag_result)
                    self.output_q.put(curr_batch)
                else:
                    print("Failed batch serialization of:\n{}\n failed on trial {} with error: {}" \
                          .format(curr_batch['upload_data'], curr_batch['retrials'], str(result)))
                    print(get_cloud_error_message(result['status_code']))
                    self.input_q.put(curr_batch)
            else:
                self.output_q.put(curr_batch)
            
            self.input_q.task_done()
            time.sleep(self.min_delay_between_batches)


if __name__ == '__main__':
    pass
